import pyproteins.container.customCollection
import pyproteins.container.Core
import re

''' Collection of Pfam entries '''

def strip(fileName):
    string = re.sub("\.[^\.]+$", "", fileName)
    return string

class EntrySet(pyproteins.container.customCollection.EntrySet, object):
    def __init__(self, collectionPath):
        if not collectionPath:
            raise ValueError("Please specify a path to pfam cache argument required")
        super().__init__(collectionPath=collectionPath, constructor=Entry, typeCheck=None, indexer=strip)
        
    def map(self, uniprotID=None):
        if uniprotID:
            e = self.get(uniprotID)
            return e.matches

class Entry(pyproteins.container.Core.Container):
    def __init__(self, id, baseUrl="http://pfam.xfam.org/protein/", fileName=None):
        if not id:
            raise TypeError('identifier is empty')
        #super(Entry, self).__init__(id, url=baseUrl + str(id) + '?output=xml', fileName=fileName)
        super().__init__(id, url=baseUrl + str(id) + '?output=xml', fileName=fileName)
        #pyproteins.container.Core.Container.__init__(self, id, url=baseUrl + str(id) + '?output=xml', fileName=fileName)
        self._parser()
    #@property
    #def xmlH(self):
    #    return self.getXmlHandler()


    def _parser(self):
        e = self.getXmlHandler()
        if not e:
            raise ValueError("not a valid Pfam xml Handler")
        if not e.find('description'):
            raise ValueError("not a valid Pfam ressource")

        self.description = re.sub("^[\n\s]*([\S]+.*[\S]+)[\n\s]*$",r"\1", e.find("description").text)

        self.matches = [ Match(x) for x in e.find_all('match') ]



#<match accession="PF00143" id="Interferon" type="Pfam-A">
#<location start="30" end="185" ali_start="31" ali_end="185" hmm_start="2" hmm_end="156" evalue="2.5e-62" bitscore="221.60"/>
#</match>
class Match():
    def __init__(self, xmlHandler):
        if xmlHandler:

            self.id = xmlHandler['id']
            self.accession = xmlHandler['accession']
            self.type = xmlHandler['type']
            self.locations = [ { k : v  for e in xmlHandler.find_all('location') for k, v in e.attrs.items()} ]

    def __repr__(self):
        string = self.id + " " + self.accession + " " + self.type + ":"
        for loc in self.locations:
            string += "{start:" + loc['start'] + " end:" + loc['end'] + "}"
        return string


# Following methods ensure duck-typing w/ Uniprot.Entry.Domain class (ie common interface)
    def owns(self, position):
        i = int (position)
        for loc in self.locations:
            if i >= int(loc['start']) and i <= int(loc['end']):
                return True
        return False

    @property
    def _dict(self):
        begin = [ int(l['start']) for l in self.locations ]
        end = [ int(l['end']) for l in self.locations ]
        return { 'description' : self.id + ":" + self.accession + ":" + self.type, 'begin' : begin, 'end' : end }



# coding: utf8
#import urllib2
from bs4 import BeautifulSoup
import re
import xml.etree.ElementTree as ET
import pyproteinsext.uniprot
import json
import os
import numpy as np
import multiprocessing
import pyproteins.container.Core as ca
PYTHONIOENCODING='utf-8'

class MitabTopology(object):
    def __init__(self, psqObject):
        self.tmpAdj = ca.dnTree(append=True)
        for psqData in psqObject:
            self.tmpAdj.append(psqData.interactors[0][0][1], 
                            psqData.interactors[1][0][1], psqData)
    
    def keys(self):
        return  list( self.tmpAdj.keys() )

    def __iter__(self):
        for d in self.tmpAdj:
            yield d

    def __len__(self):
        return len(self.tmpAdj)
            
    def __getitem__(self, k1):
        return self.tmpAdj.getNonDense(k1)

    def __repr__(self):
        return repr(self.tmpAdj)

def parse_worker(input):
    print( 'Worker^^ ' + str(len (input['bufferArray'])) )
    #for f in input:
    #    print f
    #return []
    bufferRecords = []

    looseChk = False
    if 'looseChk' in input:
        if input['looseChk'] :
            looseChk = True

    for line in input['bufferArray']:
        if len(line) == 0 or line.startswith("#"):
            continue
        if 'encoder' in input:
                #print "==>" + kwargs['encoder']
            datum = PSQDATA( line.decode(['encoder']) )
        else:
            datum = PSQDATA(line)
            #if _checkPsqData(input['self'], datum) or looseChk:
        bufferRecords.append(datum)

    return bufferRecords

def _checkPsqData (self, psqDataObj):
    pmid = psqDataObj['pmid']
    source = psqDataObj['source'].lower()
    if not pmid in self.registredPublications:        
        self.registredPublications[pmid] = source
        #print("Putting " + source +  ' in ' +  self.registredPublications[pmid])
        #print(psqDataObj)
        return True

    if self.registredPublications[pmid] == source:
        return True
    else:
        #print ("Warning publication " + pmid + " provided by " + source + " has already been fetched from " + self.registredPublications[pmid])
        #print(psqDataObj)
        return False



#from pkg_resources import resource_stream, Requirement
#

PSQ_FIELDS= ["idA", "idB", "altA", "altB", "aliasA", "aliasB", "interactionDetectionMethod", "firstAuthor", "pubid", "taxidA", "taxidB",
            "interactionTypes", "sourceDatabases", "interactionIdentifiers", "confidenceScore", "complexExpansion", "biologicalRoleA"
            , "biologicalRoleB", "experimentalRoleA", "experimentalRoleB", "interactorTypeA", "interactorTypeB", "xRefA", "xRefB",
            "xRefInteraction", "annotationA", "annotationB", "annotationInteraction", "taxidHost", "parameters", "creationDate",
            "updateDate", "checksumA", "checksumB", "negative", "featuresA", "featuresB", "stoichiometryA", "stoichiometryB",
            "identificationMethodA", "identificationMethodB"]

# BEWARE will have to move from SOAP to REST at 1st Oct 2016
# http://www.ebi.ac.uk/ols/roadmap.html
class OLS():

    def __init__(self, ontology="mi"):

        self.cOntology = ontology
        self.restApi = 'http://www.ebi.ac.uk/ols/api/ontologies/' + self.cOntology + '/'
        self.lineage = {}

    def _parse (self, url):
        response = urllib2.urlopen(self.restApi + url)
        data = json.load(response)
        return data

    def isSonOf(self, childId=None, parentId=None):
        if not childId or not parentId:
            return None
        self._getLineage(childId)
        if parentId in self.lineage[childId]: return True
        return False

    def _getLineage(self, termId):
        u = termId.replace(":", "_")
        req = self._parse("terms/http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252F" + u + "/hierarchicalAncestors")
        self.lineage[termId] = [ k['obo_id'] for k in req["_embedded"]["terms"] if  k['obo_id'] ]
        print(self.lineage[termId])

    def getTermById(self, termId=None):
        ans = self.restApi + 'terms?obo_id=' + termId
        print(ans)
        if ans == termId:
            print("id term failed")
            return None


def _convert(psqDataObj, biogridMapper):
    for i in range(0,2):
        if psqDataObj.data[i].data[0].type == 'uniprotkb:':
            continue

        tmp = str(psqDataObj.data[i].data[0])
        v = psqDataObj.data[i + 2]['biogrid']
        if not v:
            break
        u = biogridMapper(biogridId=v[0])
        if not u:
            break

        #nP = PSQFIELD('uniprotkb:' + u)
        psqDataObj.data[i] = PSQDATUM('uniprotkb:' + u)
        psqDataObj.data[i + 2].data.append(PSQFIELD(tmp))


class PSICQUIC(object):
    mitabLvls = ["25", "27"]
    olsWebService = OLS()

    def read(self, fileName, **kwargs):
        with open(fileName, 'r') as f:
           # print 'toto2'
            s = f.read()
        #return
        #if n > 1:

        #else:
            self._parseString(s, **kwargs)


    def convert(self, biogrid=None): #  bioigrid is a reference to a BIOGRID mapper object
        if biogrid:
            for psqData in self:
                _convert(psqData, biogrid)
    #returns a subset of the current PSICQUIC data, create a new PSICQUIC object
    #For now, psicquic datum are never modified, we therefore simply copy by reference
    def clone(self, **kwargs):
        cloneSelf = PSICQUIC()
        if not kwargs:
            cloneSelf.records = records
        #if 'partnerPairs' in kwargs:

    def __add__(self, other):
        self.records += [ psqDataObj for psqDataObj in other.records if _checkPsqData(self, psqDataObj) ]

    def __len__(self):
        return len (self.records)

    def __init__(self, registryUrl="http://www.ebi.ac.uk/Tools/webservices/psicquic/registry/registry?action=STATUS&format=xml", mode='LOOSE', offLine=False):
        self.records = []
        self.mode = mode # or 'STRICT' to avoid pmid from different source, less efficient than hash
        self.registredPublications = {} # { "pubid" : "sourceDatabase" }
        
        if not offLine:
            self.registry = self.getRegistry(registryUrl)
            if not self.registry:
                self.registry = registry()

    def __repr__(self):
        return  "\n".join(map(str, self.records))
        
    def __getitem__(self, i):
        return self.records[i]

    def __iter__(self):
        for record in self.records:
            yield record

    def clear(self):
        self.records = []
        self.registredPublications = {}

    def json(self, file=None):
        jsonString = '{"type" : "mitabResult", "data" : [' + ','.join([ psqData.json for psqData in self ]) + '] }'

        if file:
            with open(file, 'w') as f:
                f.write(jsonString)
        return jsonString
    def __str__(self):
        if len(self) > 100: 
            return  "100 first items...\n" + "\n".join(map(str, self.records[:100]))
        return "\n".join(map(str, self.records))
        
    def dump(self, file=None):
        if file:
            with open(file, 'w') as f:
                f.write(self.__repr__())
        #else:
        #    print self.__repr__()
        return self.__str__()
    def load(self, mitabStream):
        if not mitabStream:
            print("You must provide a mitab input")
            return
        bufferStr = ""

        for line in mitabStream:
            bufferStr = bufferStr + line
        self._parseString(bufferStr)

# Receives a 2 tuple of uniprot identifiers (h1,h2)
# Performs two requests with OR combination of w/ each set, and a AND combination of the two set
# eg : ((xxx,yyy),(uuu,vvv)) -> miql1 : idA:(xxx OR yyy) AND idB:(uuu OR vvv)
#                               miql2 : idB:(xxx OR yyy) AND idA:(uuu OR vvv)
    def zQuerySlow(self, hArray, **kwargs):
        #FS='%20'+'OR%20'
        param = kwargs
        param['erasePrevious'] = False

        for p1 in hArray[0]:
            for p2 in hArray[1]:
                if p1 == p2:
                    self.query(raw='idA:' + p1 + '%20'+'AND%20idB:' + p2, **kwargs)
                else:
                    self.query(pair=(p1, p2), **kwargs)
        #miqlStringOne = 'idA:(%20' + FS.join(hArray[0]) + '%20)%'+ '20AND%20' + 'idB:(%20' + FS.join(hArray[1]) + '%20)'
        #miqlStringTwo = 'idA:(%20' + FS.join(hArray[1]) + '%20)%'+ '20AND%20' + 'idB:(%20' + FS.join(hArray[0]) + '%20)'
        #print miqlStringOne
        #print miqlStringTwo

        #self.query(raw=miqlStringOne, **kwargs)


        #self.query(raw=miqlStringTwo, **param)

    # Must be put on two separated threads
    # Must deal w/ autointeraction
    def zQuery(self, hArray, **kwargs):
        #FS='%20'+'OR%20'
        param = kwargs
        param['erasePrevious'] = False

        cloneOne = PSICQUIC()
        cloneOne.registry

        chunks = [hArray[0][x:x+50] for x in xrange(0, len(hArray[0]), 50)]

        for p1 in chunks:
            cloneOne.query(uniprotId=p1,**kwargs)


        chunks = [hArray[1][x:x+50] for x in xrange(0, len(hArray[1]), 50)]
        cloneTwo = PSICQUIC()
        cloneTwo.registry
        for p2 in chunks:
            cloneTwo.query( uniprotId=p2, **kwargs)


        print ('**************')
        print (len (cloneOne))
        print (len (cloneTwo))

       # with open('/Users/guillaumelaunay/toto_1.txt', 'w') as f:
       #     f.write(cloneOne.dump())
       # with open('/Users/guillaumelaunay/toto_2.txt', 'w') as f:
       #     f.write(cloneTwo.dump())
        #print '####'
        #print cloneTwo.dump()

        bufRecords = set(cloneOne.records) & set(cloneTwo.records)
        self.records = list(bufRecords)
        print(len(self))

    def query(self, **kwargs):

        if 'providers' in kwargs:
            providers = kwargs['providers']
        else:
            providers = ["dip"]

        if 'mitabLvl' in kwargs:
            mitabLvl = kwargs['mitabLvl']
        else:
            mitabLvl="25"

        if 'erasePrevious' in kwargs:
            erasePrevious = kwargs['erasePrevious']
        else:
            erasePrevious = True

        if providers[0] == '*':
            providers = self.registry

        if erasePrevious:
            self.clear()

        if mitabLvl not in self.mitabLvls:
            print("invalid mitab level " + mitabLvl)
            return None

        parameterSet = ('uniprotId', 'seeds', 'pair', 'species')
        if set(parameterSet) | set(kwargs.keys()) == 0:
            raise ValueError('Missing one parameter : ' + str(parameterSet))

        miqlString = ''
        if 'raw' in kwargs:
            miqlString = kwargs['raw']
        else :
            miqlParam = []

            for k, v in kwargs.iteritems():
                if k in  ['pair', 'uniprotId', 'seeds']:
                    field = "id:"
                elif k == 'species':
                    field = "species:"
                else :
                    continue

                if k == "pair":
                    miqlParam.append(field + '(' + ('%20'+'AND%20').join(v) + ')')
                elif isinstance(v, list):
                    miqlParam.append(field + '(' + ('%20'+'OR%20').join(v) + ')')
                else :
                    miqlParam.append(field + v)

            miqlString = ('%20' + 'AND%20').join(miqlParam)

        for provider in providers:
# psicquic members
            if not provider in self.registry:
                print("provider is no registred database")
                continue
            miql = self.registry[provider] + 'query/' + miqlString
            #print miql
            ans, encoder = self._ping(miql + "?format=tab" + mitabLvl)
            if ans == 0:
                ans, encoder = self._ping(miql + "?format=tab25")
            if ans:
                self.parseString(ans, encoder=encoder)
            else:
                continue

        if not erasePrevious:
           # print "b4 NR " + str(len(self))
            self.makeNR()
           # print "AftR NR " + str(len(self))

    def _ping(self, url):
        try:
            response = urllib2.urlopen(url)
        except urllib2.HTTPError as error:
            if error.code == 406:
                print (url)
                print ("mitab Level may not be supported retrying w/ 2.5")
                return 0
            print (url + "\nHTTP ERROR " + str(error.code))
            return None
        except urllib2.URLError as error:
            print (url + "\n" + str(error.reason))
            return None
        raw = response.read()

        encoder = response.headers['content-type'].split('charset=')[1] if len(response.headers['content-type'].split('charset=')) > 1 else 'utf-8'


        #Content-Type:text/plain; charset=utf-8

        response.close()
        return (raw, encoder)

    def getRegistry(self, url):
        try:
            response = urllib2.urlopen(url)
        except urllib2.HTTPError as error:
            print ("Unable to contact registry at " + url)
            print ("Loading default")
            return registry()

        raw = response.read()
        response.close()
        return registry(raw)

    # mode = LOOSE and mprocess, encoder, have to be passed a kwargs
    def _parseString(self, raw, **kwargs):
        bufferArray = [ line for line in raw.split("\n") ]
    
        if 'n' in kwargs:
            if kwargs['n'] > 1 :
                print('Distributing parsing of ' +  str(len(bufferArray)) + ' mitab lines over ' + str(kwargs['n']) + ' processes')
                inputs =  [ { 'bufferArray' : x.tolist() }for x in np.array_split(bufferArray, kwargs['n']) ]
                    #if __name__ == '__main__':
                for d in inputs:
                    if 'encoder' in kwargs:
                        d['encoder'] = kwargs['encoder']
                    d['looseChk'] = True if self.mode is 'LOOSE' else False
                       # d['self'] = self
                pool = multiprocessing.Pool(processes=kwargs['n'])
    # map the list of lines into a list of result dict
                res = pool.map(parse_worker, inputs)
                for r in res:
                    self.records += r
                pool.close()
                pool.join()
                return

        self._parse(bufferArray)

    def _parse(self, bufferArray, **kwargs):
        
        bufferRecords = []
        for line in bufferArray:
            if len(line) == 0 or line.startswith("#"):
                continue
            if 'encoder' in kwargs:
                #print "==>" + kwargs['encoder']
                bufferRecords.append( PSQDATA( line.decode(kwargs['encoder']) ) )
            else:
                bufferRecords.append(PSQDATA(line))
        
        if self.mode is "LOOSE":
            self.records += bufferRecords
        else:
            self.records = [data for data in bufferRecords if _checkPsqData(self, data)]
        
    def analyse(self):
        if len(self) == 0: return None
        data = {
            "stats" : self.statInteractionMethods(),
            "pmids" : self.countPmid()
        }
        return data

    def countPmid(self):
        knownPmids = []
        for record in self.records:
            if record['pmid'] not in knownPmids:
                knownPmids.append(record['pmid'])
        return knownPmids

    def statInteractionMethods(self, opt=None):
        if opt :
            stats = opt
        else :
            stats = {
                "MI:0401" : { "name" : "biochemical", "count" : 0},
                "MI:0013" : { "name" : "biophysical", "count" : 0},
                "MI:0254" : {"name":"genetic interference","count" : 0},
                "MI:0428" : { "name" : "imaging technique", "count" : 0},
                "MI:1088" : { "name" : "phenotype-based detection assay", "count" : 0},
                "MI:0255" : { "name" : "post transcriptional interference", "count" : 0},
                "MI:0090" : {"name":"protein complementation assay","count":0},
                "MI:0362" : { "name" : "inference", "count" : 0},
                "MI:0063" : {"name":"interaction prediction","count":0},
                "MI:0686" : { "name" : "unspecified method", "count" : 0}
            }
        stillExperimental = 0
        for psqDataObj in self:
            detectMeth = psqDataObj["interactionDetectionMethod"]
            boolT = False
            if detectMeth in stats:
                stats[detectMeth]['count'] = stats[detectMeth]['count'] + 1
                continue
            for id in stats:
                if self.olsWebService.isSonOf(detectMeth, id):
                    stats[id]['count'] = stats[id]['count'] + 1
                    boolT = True
                    break
            if not boolT:
                if detectMeth == "MI:0045":
                    stillExperimental = stillExperimental + 1
                else:
                    print("Warning " + detectMeth + " was not cast")

        stats["experimental"] = stillExperimental
        #print stats
        #print "\tstill Experimental => " + str(stillExperimental)
        return stats



    def topology(self, type='uniprotID'): # We should register alias of interactors too
        nodes = []
        edges = {}
        for p in self:
            t = p['uniprotPair']
            if not t:
                continue
            nodes += [ n for n in t if n not in nodes ]
            if t not in edges:
                edges[t] = [p]
            else :
                edges[t].append(p)

        return (nodes,edges)

    def makeNR(self):
        self.records = list(set(self.records))


    def getBiomolecules(self, type='uniprot'):
        if type == 'uniprot':
            l = []
            for p in self:
                up = p['uniprotPair']
                if up:
                    l += up
            return list(set(l))


# Returns another instance of PSQ with filter data elements
    def filter(self, **kwargs):
        target = PSICQUIC(offLine=True)
       
        # For now we dont look into alias uniprot identifiers
        if 'uniprot' in kwargs:
            buf = kwargs['uniprot']
            if isinstance(kwargs['uniprot'], list):
                buf = set(kwargs['uniprot'])
            elif isinstance(kwargs['uniprot'], basestring):
                buf = set(kwargs['uniprot'])
#            elif isinstance(kwargs['uniprot'], set):

            for psqData in self:
                up = psqData['uniprotPair']
                if not up:
                    continue
               # print up
               # print buf
                if set(up) & buf :
                    #print str(up) + ' <===> ' + str(buf)
                    target.records.append(psqData)

        if 'predicate' in kwargs:
            f = kwargs['predicate']
            for psqData in self:
                if f(psqData):
                    target.records.append(psqData)
        
        return target

# This is the main interface we try to define smart __getitem__() access keys
class PSQDATA():
    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        
        idA = self.data[0].data[0].value
        idB = self.data[1].data[0].value
        
        max = idA if hash(idA) > hash(idB) else idB
        min = idA if hash(idA) < hash(idB) else idB
        

        return hash( min + max + self.data[1].data[0].value + self["interactionDetectionMethod"] + self["pmid"])

    def __init__(self, raw):
        self.raw = raw
        self.data = [PSQDATUM(column) for column in re.split(r'\t+', raw)]#raw.split("\t") ] #if column

        if len(self.data) != 15 and len(self.data) != 42:
            for i,e in enumerate(self.data):
                print ('[' + str(i) + '] ' + str(e))
            raise ValueError ("Uncorrect number of tabulated fields on input [" +
                str(len(self.data)) + "] at:\n" + str(raw) )

    def __repr__(self):
        string = "\t".join(map(str,self.data))
        return string
    def __getitem__(self, key):
        if key is "taxid":
            return (self.data[9].content[0][1], self.data[10].content[0][1])
        if key is "pmid":
            for field in self.data[8].data:
                if field.type == "pubmed:":
                    return field.value
            return self.data[8].data[0].value
        if key is "source":
            if self.data[12].data[0].annotation:
                return self.data[12].data[0].annotation
            return self.data[12].data[0].value
        if key is "interactionDetectionMethod":
            return self.data[6].data[0].value
        if key is "species":
            return (self.data[9].data[0].value, self.data[10].data[0].value)
        if key is 'uniprotPair':
            a = pyproteinsext.uniprot.capture(self.data[0].data[0].value) if pyproteinsext.uniprot.capture(self.data[0].data[0].value) else pyproteinsext.uniprot.capture(self.data[2].data[0].value)
            b = pyproteinsext.uniprot.capture(self.data[1].data[0].value) if pyproteinsext.uniprot.capture(self.data[1].data[0].value) else pyproteinsext.uniprot.capture(self.data[3].data[0].value)
            if a and b:
                (a,b) = (b,a) if b < a else (a,b)
                return (a, b)
            return None


    @property
    def json(self):
        return json.dumps({ k : str(d)for k,d in zip(PSQ_FIELDS, self.data) })

    @property
    def interactors(self):
        datum = (self.data[0].content + self.data[2].content, self.data[1].content + self.data[3].content )
        #print datum
        return  datum
    
    def swapInteractor(self, to, iSlot=None): # iSlot to force, maybe autoassociation ?
        consideredSlots = [0, 1]
        if iSlot:
            try :
                consideredSlots= [ ['A', 'B'].index(iSlot) ]
            except ValueError:
                print("if you specify a slot to swap interactors it must be \"A\" or \"B\"")
                return

        for i in consideredSlots:
            psqDatumObj = self.data[i]
            psqDatumObjAlt = self.data[i + 2]
            #print (str(i) + '::' + str(psqDatumObjAlt))
            for iField, cPsqField in enumerate(psqDatumObjAlt):
             #   print(cPsqField.value)
                if cPsqField.value == to:                    
                    _psqFieldAnon = cPsqField
                    psqDatumObjAlt.data[iField] = psqDatumObj.data[0] # Access the 1st psqField
                    psqDatumObj.data[0] = _psqFieldAnon
                    break

    def hasInteractors(self, mode='STRICT'):
    #    for psqDatum in [ self.data[0].content + self.data[]
        pass

    def getNames(self):
        pass

    def getPartners(self):
        pass
            # Ask for partners
            # Extract uniprot id
            # fill a 'p->{ m_0, m_1, ..., m_n,}, where m's are uniprot match

class PSQDATUM():
    def __init__(self, string):
        self.raw = string
        self.data = [ PSQFIELD(field) for field in string.split('|')]
    def __repr__(self):
        string = '|'.join(map(str,self.data))
        return string

    def __iter__(self):
        for psqFieldObj in self.data:
            yield psqFieldObj

    def __getitem__(self, key):
        hits = []
        for psqFieldObj in self.data:
            if psqFieldObj.type == key + ':':
                hits.append(psqFieldObj.value)
        return hits


    @property
    def content(self):
        return [ (e.type, e.value) for e in self.data ]



class PSQFIELD():
    fieldParser = re.compile('^([^:^"]+:){0,1}"{0,1}([^"\(]+)"{0,1}\({0,1}([^\)]+){0,1}\){0,1}$')
    def __init__(self, raw):
        m = re.match(self.fieldParser, raw)
        self._raw = raw
        if not m:
            #print "warning following field causes problem to parse\n" + raw
            self.value = raw
            self.type = None
            self.annotation = None
        else:
            self.type = m.groups()[0]
            self.value = m.groups()[1]
            self.annotation = m.groups()[2]
   # def __repr__(self):
        #if isinstance(self.value, basestring):
   #     if not isinstance(self.value, unicode):
    #        string = unicode(self.value, 'utf-8')
    #    else :
    #        string = self.value

    #    if self.type: string = self.type + string
    #    if self.annotation: string = string + "(" + self.annotation + ")"
    #    return string

    def __str__(self):
        return self._raw

        try :
            string = self.value.decode('ascii')
        except UnicodeDecodeError:
        #    print "Decode error" + self.value
            string = self.value.decode('utf8')
        except UnicodeEncodeError:
            #print "UTF ENCODING"
            string = self.value.encode('utf8')
        #string = self.value

        if self.type is 'comment':
            return self.raw

        if self.type:
            string = self.type + string
        if self.annotation:

            try :
                str_annot = self.annotation.decode('ascii')
            except UnicodeDecodeError:
                #print "OUPS ANNNOT"
                str_annot = self.annotation.decode('utf8')
            except UnicodeEncodeError:
                #print "OUPSS EEECC"
                str_annot = self.annotation.encode('utf8')
        #    string = string + "(" + str(str_annot) + ")"
            string = string + "(" + str(self.annotation) + ")"
        else:
            pass

        return string



class registry():
    data = {
        'dip' : "http://imex.mbi.ucla.edu/psicquic-ws/webservices/current/search/",
        'intact' : "http://www.ebi.ac.uk/Tools/webservices/psicquic/intact/webservices/current/search/",
        'mint' : "http://www.ebi.ac.uk/Tools/webservices/psicquic/mint/webservices/current/search/",
        'innatedb_imex' : "http://www.ebi.ac.uk/Tools/webservices/psicquic/innatedb/webservices/current/search/",
        'matrixdb' : "http://matrixdb.ibcp.fr:8080/psicquic/webservices/current/search/",
        'innatedb' : "http://psicquic.curated.innatedb.com/webservices/current/search/"
    }

    def __init__(self, raw):
        if raw:
            #print raw
            try:
                root = ET.fromstring(raw)
            except:
                print("Registry Remote XML parsing error, using local file as registry")
#                resource_stream(Requirement.parse("pyproteinsExt=="), "restez/httpconn.py")

                path =  os.path.abspath(__file__)
                dir_path = os.path.dirname(path)
                tree = ET.parse(dir_path + '/static/psicquicRegistryDefault.xml')
                root = tree.getroot()
            #else:


            for child in root:
                name = ""
                url = ""
                for subT in child:
                    if subT.tag == "{http://hupo.psi.org/psicquic/registry}restUrl":
                        url = subT.text
                    if subT.tag == "{http://hupo.psi.org/psicquic/registry}name":
                        name = subT.text
                name = name.lower().replace("-","_")
                self.data[name] = url

                '''
                if not line:
                    continue
                #print line
                field = line.split("=")
                name = field[0].lower().replace("-","_")
                self.data[name] = field[1]
                '''

    def __getitem__(self,index):
        if not index in self.data:
            return None
        return self.data[index]

    def __repr__(self):
        string = ""
        for db in self.data:
            string = string + db + " : " + self.data[db] + "\n"
        return string

    def __iter__(self):
        for db in self.data:
            yield db



# This package provides utility function to manipulate the GO/MI ontology terms
# check performed on passed string is a bit clunky
# "obo:MI_0090" or "MI:0090" are OK
import re
from owlready2 import *


def isOboRegular(term):
    # print "Coucou " + term
    # print re.match(r'^obo:[A-Z]{2}:[0-9]{4}$', term)
    return re.match(r'^obo:[A-Z]{2}_[0-9]{4}$', term)


def isOboNamespaced(term):
    return re.match(r'[^:]+:[0-9]+', term)


# Manipulating the MolecularInteraciton ontology
# print model.classes
# aClass = model.getClass("MI_0090")
# aClass[0].children()
# model.printClassTree()


class Ontology():
    def __init__(self, owlFile=None, url=None, oType='GO'):
        self.type =oType
        # super(Ontology, self).__init__(ressource)
        if not owlFile is None :
            self.onto = get_ontology("file://" + owlFile).load()
        elif not url is None:
            self.onto = get_ontology(url).load()
        else:
            raise IOError("No ontology source specifed!")
    
    # find terms based on namespaced Id or text description in label
    def find(self, stringTerm):
        if isOboNamespaced(stringTerm):
            elem = self.onto.search(id=stringTerm)            
            if not elem:               
                elem = self.onto.search(hasAlternativeId=stringTerm)
            if not elem:
                raise KeyError(f"No element \"{stringTerm}\"")
            return elem
        return self.onto.search(label=stringTerm)

    def findOne(self, stringTerm):
    #    if isOboNamespaced(stringTerm):
    #        return self.onto.search_one(id=stringTerm)
    #    return self.onto.search_one(label=stringTerm)
        if isOboNamespaced(stringTerm):
            elem = self.onto.search_one(id=stringTerm)
            if not elem:
                #print("T2")
                elem = self.onto.search_one(hasAlternativeId=stringTerm)
            if not elem:
                #print("T3")
                raise KeyError(f"No element \"{stringTerm}\"")
            return  elem
        return self.onto.search_one(label=stringTerm)

    # if term is a/list of string coherce it into matching ontology classes
    def _coherceIntoMany(self, data):
        if not isinstance(data, list):
            data = [data]
        res = []
        for e in data:
            res += self.find(e)
        return list(set(res))

    # Given a list of term of termlike, (including regexp)
    # Get the non redundant list of terms that match the supplied argument and all their sons

    def harvest(self, termLikeList):

        seedNodes = self._coherceIntoMany(termLikeList)
        print('Total number of nodes matching request : ' + str(len(seedNodes)))

        secondaryNodes = []
        for seedNode in seedNodes:
            secondaryNodes += self.onto.search(subclass_of=seedNode)
        secondaryNodes = list(set(secondaryNodes))
        print('Total number of potential nodes found under seed nodes : ' + str(len(secondaryNodes)))

        return list(set(seedNodes + secondaryNodes))

    def isSonOf(self, x, y):
        x, y = self._coherce([x, y])
        return y in self._getLineage(x)

    def _getLineage(self, term):
        term = self._coherce(term)[0]

        lineage = []

        tmpNode = self._rollupNode(term)
        if not tmpNode:
            print(str(term) + ' has no parent')

        while tmpNode:
            lineage.append(tmpNode)
            tmpNode = self._rollupNode(tmpNode)

        return lineage

    @staticmethod
    def _rollupNode(e):
        if not e.is_a:
            return None
        return e.is_a[0]

    # Given a list of terms regroup them in clusters based on the member list
    # that are the highest in the hierarchy
    # INCOMPLETE
    def cluster(self, termsList):

        def _hasAnyDad(e, l):
            for _e in l:
                # print(str(_e) + ' ' + str(e))
                if self.isSonOf(e, _e):
                    return True
            return False

        # Create a list of term that have no parents

        rootNodes = [{e: []} for e in termsList if not _hasAnyDad(e, termsList)]

        return rootNodes

    # Find and Count foreach provided term in
    # the correesponding parents in range domain the number
    # Not tested w/ owlReady !!!
    def project(self, domainTerms, rangeTerms, flatDic=False):

        domainTerms = self._coherce(domainTerms)
        res = {x: [] for x in self._coherce(rangeTerms)}
        others = []
        for query in domainTerms:
            known = False
            for rangeTerm in res:
                if self.isSonOf(query, rangeTerm):
                    res[rangeTerm].append(query)
                    known = True
            if not known:
                others.append(query)

        if not flatDic:
            res["others"] = others
            return res

        data = {"term": [], "parent": []}
        for rangeTerm in res:
            for son in res[rangeTerm]:
                data["term"].append(str(son.bestLabel()))
                data["parent"].append(str(rangeTerm.bestLabel()))
        for query in others:
            data["term"].append(str(query.bestLabel()))
            data["parent"].append("others")

        return data

    # if term is a string coherce it in a ontology class object
    def _coherce(self, data):
        if not isinstance(data, list):
            data = [data]
        return [self._termCoherce(e) for e in data]

    def _termCoherce(self, termLike):
        if isinstance(termLike, str):
            term = self.findOne(termLike)
            if not term:
                raise ValueError("No such term found : \"" + str(termLike) + "\"")
            termLike = term

        return termLike

    def constructTreeFromRoot(self, root_id: str, things_to_add: dict = None):
        """
        Construct a Tree instance from root_id node (id=root_id).
        Returns a Tree object.

        things_to_add may be an iterable of string specified which attribute of the owl node should be imported with.
        Default imported attributes are id and label, stored in node.id and node.label.
        Every other key you need to add should be specified in things_to_add. An inexistant property in a node will be ignored.
        Data will be available in node.misc, in a dictionnary.

        things_to_add may be a function taking a OwlNode in parameter (also known as ThingClass) and returning None OR a list of string.
        List of string will be treated for this node like specified in the previous paragraph.

        Examples:
            # Classic

            my_tree = ontology.constructTreeFromRoot("MI:0001")

            # Let's adding some additionnals informations to the nodes

            # With iterables

            my_tree = ontology.constructTreeFromRoot("MI:0001", ['hasExactSynonym'])

            print(my_tree.root.misc['hasExactSynonym'])  # prints "interaction detect"

            # With functions

            my_tree = ontology.constructTreeFromRoot("MI:0001", lambda x: ['hasExactSynonym'] if len(x.label) > 1 else None)
        """

        def find(predicate, iterable):
            for e in iterable:
                if predicate(e):
                    return e
            return None

        def generateUntilSeed(predicate, iterable):
            for e in iterable:
                yield e
                if predicate(e):
                    break

        def getThingsToAdd(node):
            things = things_to_add

            if callable(things_to_add):
                things = things_to_add(node)

            if things:
                d = {}

                for k in things:
                    try:
                        d[k] = getattr(node, k)
                    except AttributeError:
                        # Does not exists
                        pass

                if len(d):
                    return d

            return None

        nodes = self.harvest(root_id)

        tree = Tree()

        rootnode = self.onto.search_one(id=root_id)

        if not rootnode:
            raise IndexError("Seed does not exists")

        tree.root = Node(rootnode.id[0], rootnode.label[0], misc=getThingsToAdd(rootnode))

        for c in nodes:
            # print(c, c.id, c.label)
            try:
                c_id, c_label, c_misc = c.id[0], c.label[0], getThingsToAdd(c)

                lineage = map(
                    lambda x: (x.id[0], x.label[0], getThingsToAdd(x)) if x.name != 'Thing' else ("__root__", None),
                    self._getLineage(c_id))

                # Filtre lineage jusqu'à trouver le root. Arrête ensuite
                lineage = list(generateUntilSeed(lambda x: x[0] == root_id, lineage))

                if not find(lambda x: x[0] == root_id, lineage):
                    continue

                tree.append(lineage, c_id, c_label, c_misc)
            except IndexError:
                pass

        return tree


# Needed for create tree for Ontology.constructTreeFromRoot()
class Tree:
    """Tree with unique nodes, using storage containing only IDs and labels."""

    def __init__(self):
        self.root = None
        self.node_dict = {}

    def toDict(self):
        """
        Get dictionnary version of a tree.

        Supposing an interface like:

        interface NodeChild { name: str /* Node label */, children: { [childId: str]: NodeChild }, misc?: any }


        Dictionnary will be organised like:


        {
            root_node_id: NodeChild
        }
        """
        return {
            self.root.id: self.root.toDict()
        }

    def append(self, parents, current_id, current_label, misc=None):
        """Append a node to the tree.

        Parents must be a list of tuples containing in [0]: id, [1]: label. Ordered by child proximity, the first one is the nearest.
        It can be a "misc" dict in [2] position.
        """
        # Si le noeud existe déjà
        if self.findInTree(current_id):
            return

        i = 0
        for p in parents:
            parent_id = p[0]
            parent_l = p[1]
            parent_misc = p[2] if len(p) > 2 else None

            if self.findInTree(parent_id):
                break

            self.append(parents[i + 1:], parent_id, parent_l, parent_misc)

            i += 1

        new_node = Node(current_id, current_label, misc=misc)

        if len(parents) == 0:
            self.root.parent = new_node
            new_node.addChild(self.root)
            self.root = new_node
        else:
            parent = self.findInTree(parents[0][0])
            # print("parent", parents, parents[0][0], parent, self.root.children)
            new_node.parent = parent
            parent.addChild(new_node)

        self.node_dict[new_node.id] = new_node

    def findInTree(self, id: str):
        if self.root is None:
            return None

        if id in self.node_dict:
            return self.node_dict[id]

        return self.root if self.root.id == id else self.root.findInNode(id)

    def clone(self, deep=True):
        new_tree = Tree()
        new_tree.root = self.root.clone(deep=deep)

        return new_tree

    def prune(self, seeds: list):
        new_tree = self.clone()

        # Prune new tree with specific seeds
        okay = new_tree.root.prune(set(seeds))

        if not okay:
            raise IndexError("Cant find any node matching given seeds")

        return new_tree

    def toEte3(self):
        return self.root.toEte3()


class Node:
    def __init__(self, id, label, parent=None, misc=None):
        self.parent = parent
        self.id = id
        self.label = label
        self.children = {}
        self.misc = misc

    def addChild(self, child):
        if not isinstance(child, Node):
            raise TypeError('Must be Node type')

        child.parent = self
        self.children[child.id] = child

    def removeChild(self, id):
        del self.children[id]

    def childExists(self, id):
        return id in self.children

    def childrens(self):
        return self.children.values()

    def isLeaf(self):
        return self.children.__len__() == 0

    def toDict(self):
        d = {
            'name': self.label,
            'children': {x.id: x.toDict() for x in self.childrens()}
        }

        if self.misc:
            d['misc'] = self.misc

        return d

    def findInNode(self, id: str):
        if id in self.children:
            return self.children[id]

        for c in self.childrens():
            n = c.findInNode(id)

            if n:
                return n

        return None

    # Need to import ete3 and give TreeNode to this func
    def toEte3(self, TreeNode):
        node = TreeNode(name=self.id)
        node.add_feature("label", self.label)

        for c in self.childrens():
            node.add_child(c.toEte3())

        return node

    def clone(self, deep=True):
        # misc will be swallow copied
        cur = Node(self.id, self.label, misc=self.misc)

        if deep:
            for c in self.childrens():
                child_clone = c.clone(deep=deep)
                cur.addChild(child_clone)

        return cur

    # seeds is an iterable, ideally a set
    def prune(self, seeds):
        one_node_is_okay = False

        children_to_remove = []

        for e in self.children:
            okay = self.children[e].prune(seeds)

            if not okay:
                children_to_remove.append(e)
            else:
                one_node_is_okay = True

        for c in children_to_remove:
            self.removeChild(c)

        if not one_node_is_okay:
            one_node_is_okay = self.id in seeds

        return one_node_is_okay



import re
import pyproteinsext.uniprot
import copy
import math
import pyproteinsext.matrisome
from queue import Queue
from threading import Thread
from collections import OrderedDict
import pandas as pd
import numpy as np

import copy

# Basic input is a pdFrame, CSV maybe

# Takes an iterable, along with a coherce to uniprot Identifier
# returns and list of tuple with each iterable element and its uniprot Object
#
# Iterable ---->(Iterable, Uniprot(reduce(Iterable))

class pandaUniprotEnricher(object):
    def __init__(self, pdFrame=None, lambdaFn=None, **kwargs):
        self.pd = pdFrame
        if 'n' in kwargs:
            self.pd = pdFrame[:n]
        self.lambdaFn = lambdaFn
        if 'uniAno' in kwargs:
            self.data = kwargs['uniAno']
        else:
            self.data = [ UniAno(lambdaFn(e)) for index, e in self.pd.iterrows() ]
            print("Pooling")
            _threadPool(self.data);

        noUniprotBound = []
        _data = []
        for i,e in enumerate(self.data):
            if not e._uniprotBound:
                noUniprotBound.append(i)
                continue
            _data.append(e)

        if noUniprotBound:
            print ("Following table elements will be dropped due to unbound uniprot object")
            print ( str([self.data[n].uniprotID for n in noUniprotBound ]) )
        #print (noUniprotBound)
        #for n in reversed(noUniprotBound):
        #    print ("Droping table element " + str(n) + " (" + self.data[n].uniprotID + ") due to unbound uniprot object")
        df = self.pd.drop(self.pd.index[noUniprotBound]).reset_index(drop=True)
        self.pd = df
        self.data = _data  

    def __len__(self):
        if len(self.data) != len(self.pd):
            raise ValueError('data length dont match')
        return len(self.data)


# We select out of the padaframe only entries that match annotator
#https://chrisalbon.com/python/data_wrangling/pandas_selecting_rows_on_conditions/

# Create a dataframe
#raw_data = {'first_name': ['Jason', 'Molly', np.nan, np.nan, np.nan],
#        'nationality': ['USA', 'USA', 'France', 'UK', 'UK'],
#        'age': [42, 52, 36, 24, 70]}
#df = pd.DataFrame(raw_data, columns = ['first_name', 'nationality', 'age'])
#df

# Merging dataframes
#https://pandas.pydata.org/pandas-docs/stable/merging.html

    def filter(self, annotator, collapseCol=True, exclude=True):

        _annot, status = annotator.annotateAll(self.data)
        # fix a column order
        colHeadOrdered = [ term for term in _annot ]
        # stringify terms occurences


        #print (_annot)


        for term in _annot:
            _annot[term] = [ ','.join([ str(tOcc) for tOcc in terms ]) for terms in _annot[term] ]
       
        #return _annot
        df = self.pd
        uniAnoList =[]
        # reduce list of annotations and original dataframe
        if exclude:
            annot = {}
            for term in _annot:
                annot[term] = [ e for i,e in enumerate(_annot[term]) if status[i] ]
            print ('Getting rid of ' + str(status.count(False)) + ' entries')
            df = df.drop(df.index[ [ i for i,e in enumerate(status) if not e ] ])
            uniAnoList = [ self.data[i] for i,e in enumerate(status) if e ]
            
            #return df
        else:
            annot = _annot
            uniAnoList = self.data

        if collapseCol:
           
            _df = pd.DataFrame(annot, columns = colHeadOrdered)
            df.reset_index(drop=True, inplace=True)
            _df.reset_index(drop=True, inplace=True)

         
            df = pd.concat([df, _df], axis=1)
         
        if exclude:
            df = df.replace('', np.nan, regex=True).dropna(axis=1,how='all').replace(np.nan, '', regex=True)
        return pandaUniprotEnricher(pdFrame=df, lambdaFn = self.lambdaFn, uniAno=uniAnoList)
# panda integer indexing
# https://stackoverflow.com/questions/16096627/selecting-a-row-of-pandas-series-dataframe-by-integer-index
    def __getitem__(self,i):
        return ( self.data[i], self.pd.iloc[[i]] )

class UniAno(object):
    def __init__(self, _id):        
        self.uniprotID = _id
        self._uniprotBound = None

    #def __str__(self):
    #    return str(self._uniprotBound);

    def _boundUniprot(self):
        if not self._uniprotBound:
            id = self.uniprotID

            if not id:
                return False
            try :
                notIsoform_id = pyproteinsext.uniprot.strip(id)
                self._uniprotBound = uniprotEntrySet.get(notIsoform_id)#Uniprot.Entry(id)
            except TypeError as msg:
                print ("Following element could not be bound to uniprot entity, reason " + str(msg))
                print (self.uniprotID)
                return False
            except ValueError as msg:
                print ("Following element could not be bound to uniprot entity, reason " + str(msg))
                print (self.uniprotID)
                return False
        return True

def _threadPool(stack):
    num_worker_threads = 10
    def worker():
        while True:
            e = q.get()
            e._boundUniprot()
            q.task_done()

    q = Queue()
    for i in range(num_worker_threads):
        t = Thread(target=worker)
        t.daemon = True
        t.start()

    for e in stack:
        #print("Pouet")
        q.put(e)


    print ('*** Main thread waiting ' + str(q.qsize()))
    q.join()
    print ('*** Done')






import pyproteinsext.proteinContainer
import re
import gzip


def parse(inputFile=None):
    """
    Parse tmhmm output to create Container

    :param inputFile: path to tmhmm output
    :return: tmhmmContainerFactory.Container
    """
    mainContainer = Container()
    fnOpen = open
    t = 'r'
    if inputFile.endswith('gz'):
        fnOpen = gzip.open
        t = 'rt'
    try:
        f = fnOpen(inputFile, t)
    except IOError:
        print("Could not read file:", inputFile)
        return Container()

    instance = []
    with f:
        for l in f:
            if l.startswith("#") and "Length" in l:
                if instance:
                    mainContainer = mainContainer.addParsing(Container(input=instance))
                instance = []
            instance.append(l.rstrip())
        mainContainer = mainContainer.addParsing(Container(input=instance))
    return mainContainer


def _parseBuffer(input):
    def parseFragments(list_fragments):
        fragmentsObj = []
        reLocation = re.compile("[\s]+([\d]+)[\s]+([\d]+)")
        for f in list_fragments:
            f_split = f.split("\t")
            cellular_location = f_split[2]
            start = f_split[3].split(" ")
            location = reLocation.findall(f_split[3])[0]
            start = int(location[0])
            end = int(location[1])
            fragment = Fragment(cellular_location, start, end)
            fragmentsObj.append(fragment)
        return fragmentsObj

    prot = input[0].split(" ")[1]
    nb_helix = [l for l in input if "Number of predicted TMHs" in l][0].split(":")[1].strip()
    prot_length = [l for l in input if "Length" in l][0].split(":")[1].strip()
    fragments = parseFragments([l for l in input if not l.startswith("#")])
    obj = TMHMM_Obj(prot, int(prot_length), int(nb_helix), fragments)
    return {prot: obj}


class Container(pyproteinsext.proteinContainer.Container):
    """ Iterable container, that stores TMHMM_Obj objects """
    def __init__(self, input=None):
        super().__init__(_parseBuffer, input)


class TMHMM_Obj():
    """
    TMHMM_Obj object stores informations about one TMHMM entry. 

    :param prot: Protein id
    :type prot: str
    :param prot_length: Protein length
    :type prot_length: int
    :param nb_helix: Number of predicted transmembrane helix by TMHMM
    :type nb_helix: int
    :param fragment: List of proteins fragments (inside, outside or helix)
    :type fragments: Fragment[]
    """

    def __init__(self, prot, prot_length, nb_helix, fragments):
        self.prot = prot
        self.prot_length = prot_length
        self.nb_helix = nb_helix
        self.fragments = fragments

    @property
    def topology_seq(self):
        """
        Protein sequence with 'o' for outside loop, 'i' for inside loop and helix number for helixes

        .. warning:: obsolete if we have more than 9 helixes
        """
        # /!\ WARNING : obsolete if we have more than 9 helixes !!! Change that.
        topology_seq = ["*"]*self.prot_length
        helix_number = 1
        for f in self.fragments:
            if f.cellular_location == 'inside':
                letter = "i"
            elif f.cellular_location == "outside":
                letter = "o"
            elif f.cellular_location == "TMhelix":
                letter = str(helix_number)
                helix_number += 1
            for i in range(f.start-1, f.end):
                topology_seq[i] = letter
        if "*" in topology_seq:
            print("WARNING : * in topology seq")
        return "".join(topology_seq)


class Fragment():
    """
    :param cellular_location: TMHMM cellular location of fragment (inside|outside|TMHelix)
    :type cellular_location: str
    :param start: fragment start position
    :type start: int
    :param end: fragment end position
    :type end: int
    """

    def __init__(self, cellular_location, start, end):
        self.cellular_location = cellular_location
        self.start = start
        self.end = end



class Container(object):
    def __init__(self, fParsing=None, input=None):
        if not input:
            self.entries = {}
        else:
            self.entries = self.parsing(input, fParsing)

    def addParsing(self, other):
        for k in other.entries:
            if k not in self.entries:
                self.entries[k] = other.entries[k]
        return self

    def addEntry(self, new_entry):
        if new_entry.prot not in self.entries:
            self.entries[new_entry.prot] = new_entry

    def __len__(self):
        return len(self.entries)

    def __iter__(self):
        for protein in self.entries:
            yield self.entries[protein]

    def parsing(self, input, fParsing):
        return fParsing(input)



import pyproteins.services.utils
class Matrisome:
    def __init__(self, masterFile=None):
        if not masterFile:
            raise ValueError ("You must provide a masterFile argument to Matrisome constructor")

        tsvBuffer = pyproteins.services.utils.tsvToDictList(fileName=masterFile)
        self.data = tsvBuffer['data']
        self.keymap = tsvBuffer['keymap']
        self.accessors = self._index()

    # index entry using relevant key found in their record
    def _index(self):
        accessors = {'uniprot' : {}, 'gene' : {}} # any other meaningfull accessor
        for d in self.data:
            #print "-->" + d['UniProt_IDs'] + '<--'
            if d['UniProt_IDs']:
                for k in d['UniProt_IDs'].split(':'):
                    if k in accessors['uniprot']:
                        print (k + ' found multiple time')
                    else :
                        accessors['uniprot'][k] = []
                    accessors['uniprot'][k].append(d)
            if d['Gene Symbol']:
                for k in d['Gene Symbol'].split(':'):
                    if k in accessors['gene']:
                        print (k + ' found multiple time')
                    else :
                        accessors['gene'][k] = []
                    accessors['gene'][k].append(d)
        return accessors

    def get(self, uniprotID=None):
        if uniprotID:
            if not uniprotID in self.accessors['uniprot']:
                return []
            return [{'Category' : d['Category'], 'Division' : d['Division']} for d in self.accessors['uniprot'][uniprotID]]
        return []



import pyproteinsext.hmmrContainerFactory as hmmr
import pyproteinsext.tmhmmContainerFactory as tmhmm
import pyproteinsext.fastaContainerFactory as fasta
import pyproteinsext.proteinContainer
from collections import OrderedDict
from ete3 import NCBITaxa
from statistics import mean
from igraph import Graph
import copy


def check_if_same_proteins(dic_container):
    hmmr_proteins = set()
    tmhmm_proteins = set()
    fasta_proteins = set()
    if dic_container['hmmr']:
        hmmr_proteins = set([h.prot for h in dic_container['hmmr'].hmmrEntries])
    if dic_container['tmhmm']:
        tmhmm_proteins = set([e.prot for e in dic_container['tmhmm']])
    if dic_container['fasta']:
        fasta_proteins = set([e.prot for e in dic_container['fasta']])

    list_proteins = [hmmr_proteins, tmhmm_proteins, fasta_proteins]
    if len(list_proteins) <= 1:
        return True
    for i in range(len(list_proteins)):
        for j in range(i+1, len(list_proteins)):
            prot1 = list_proteins[i]
            prot2 = list_proteins[j]
            if len(prot1) != len(prot2):
                return False
            if prot1.difference(prot2):
                return False
    return True


def parse(hmmrOut, tmhmmOut, fastaOut):
    container = TopologyContainer()
    hmmrContainer = hmmr.parse(hmmrOut)
    tmhmmContainer = tmhmm.parse(tmhmmOut)
    fastaContainer = fasta.parse(fastaOut)

    dic_container = {'hmmr': hmmrContainer, 'tmhmm': tmhmmContainer, 'fasta': fastaContainer}
    if not check_if_same_proteins(dic_container):
        raise Exception("not same proteins ", hmmrOut, tmhmmOut, "Check !")
    container.addParsing(TopologyContainer(input=dic_container))
    return container


def _parseBuffer(dic_container):
    hmmrContainer = dic_container['hmmr']
    tmhmmContainer = dic_container['tmhmm']
    fastaContainer = dic_container['fasta']
    dic_obj = {}
    for f in fastaContainer:
        p = f.prot
        hmmr = [h for h in hmmrContainer.hmmrEntries if h.prot == p]
        tmhmm = tmhmmContainer.entries[p]
        fasta = fastaContainer.entries[p]
        obj = Topology(p, hmmr, tmhmm, fasta)
        dic_obj[p] = obj
    return dic_obj


class TopologyContainer(pyproteinsext.proteinContainer.Container):
    def __init__(self, input=None, ete3_tree=None, domain_entries=None):
        super().__init__(_parseBuffer, input)
        self.ete3_tree = ete3_tree
        self.domain_entries = domain_entries

    def filter(self, fPredicat, **kwargs):
        new_container = TopologyContainer()
        for e in self:
            if fPredicat(e, **kwargs):
                new_container.addEntry(e)
        return new_container

    def __getitem__(self, index):
        return list(self.entries.values())[index]

    def get_domain_mfasta(self, domain):
        mfasta = ''
        for e in self:
            for hit in e.hmmr:
                if hit.domain == domain:
                    header = ">" + hit.prot + " " + hit.domain
                    seq = self.get_seq(hit.hit)
                    mfasta += header + "\n" + seq + "\n"
        return mfasta

    def get_seq(self, hit):
        seq = hit.aliStringLetters
        seq = seq.replace("-", "")
        seq = seq.upper()
        return seq

    def proteins_mfasta(self):
        mfasta = ''
        for e in self:
            mfasta += ">" + e.fasta.header + "\n" + e.fasta.seq + "\n"
        return mfasta

    def complete_hmmr(self, hmmscan_out):
        container = hmmr.parse(hmmscan_out)
        new_proteins = set([h.prot for h in container.hmmrEntries])
        self_proteins = set(self.entries.keys())
        if len(new_proteins) != len(self_proteins):
            raise Exception("full Pfam proteins and original proteins are not the same")
        else:
            if new_proteins.difference(self_proteins):
                raise Exception("full Pfam proteins and original proteins are not the same")
        for e in self:
            hits = [h for h in container.hmmrEntries if h.prot == e.prot]
            e.hmmr += hits

    def filter_hit(self, fPredicat, **kwargs):
        new_container = TopologyContainer()
        for e in self:
            add = False
            hit_to_add = []
            for hit in e.hmmr:
                if fPredicat(hit, **kwargs):
                    add = True
                    hit_to_add.append(hit)
            if add:
                new_e = Topology(e.prot, hit_to_add, e.tmhmm, e.fasta, e.taxo,
                                 e.uniprot_entry,
                                 e.annotated_domains_fragments,
                                 e.Nter_UR_fragment,
                                 e.Cter_UR_fragment, e.helix_fragments,
                                 e.loop_fragments)
                new_container.addEntry(new_e)
        return new_container

    def filter_last_helix(self, distance=90):
        new_container = TopologyContainer()
        c = 0
        for e in self:
            new_helix_fragments = copy.deepcopy(e.helix_fragments)
            new_loop_fragments = copy.deepcopy(e.loop_fragments)
            if len(e.helix_fragments) == 7:
                c += 1
                new_helix_fragments = new_helix_fragments[:-1]
                new_loop_fragments = new_loop_fragments[:-1]
            elif e.helix_fragments[-1]["start"]-e.helix_fragments[-2]["end"] > distance:
                c += 1
                new_helix_fragments = new_helix_fragments[:-1]
                new_loop_fragments = new_loop_fragments[:-1]
            new_e = Topology(e.prot, e.hmmr, e.tmhmm, e.fasta, e.taxo,
                             e.uniprot_entry,
                             e.annotated_domains_fragments,
                             e.Nter_UR_fragment, e.Cter_UR_fragment,
                             new_helix_fragments, new_loop_fragments,)
            new_container.addEntry(new_e)
        print(c, "proteins have been filtered")
        return new_container

    def compute_overlapped_domains(self, overlap_accept_size):
        self.reinitialize_overlapped_domains()
        for e in self:
            for i in range(len(e.hmmr)):
                for j in range(i+1, len(e.hmmr)):
                    hit1 = e.hmmr[i]
                    hit2 = e.hmmr[j]
                    if hit1.is_overlapping(hit2, overlap_accept_size):
                        hit1.overlapped_hits.append(hit2)
                        hit2.overlapped_hits.append(hit1)

    def reinitialize_overlapped_domains(self):
        for h in [h for e in self for h in e.hmmr]:
            h.reinitialize_overlapped_hits()

    def create_ete3_tree(self):
        ncbi = NCBITaxa()
        taxids = set([e.taxo.taxid for e in self])
        if None in taxids:
            raise Exception("Entries doesn't have taxids")
        tree = ncbi.get_topology(list(taxids))

        # Complete Tree object with list of domains and proteins for each node
        node_list = []
        for n in tree.traverse('postorder'):  # Browse tree in postorder, starts from leaf and ascend to root
            n.sameDomainNode = set()
            node_list.append(n)
            n.domains = set([h.domain for e in self for h in e.hmmr if e.taxo.taxid == n.name])
            n.proteins = set([e.prot for e in self if e.taxo.taxid == n.name])
            if n.get_descendants():
                for child in n.children:
                    n.domains.update(child.domains)
                    n.proteins.update(child.proteins)

        # Complete Tree object with list of nodes with same domains for each node
        c = 0
        for i in range(len(node_list)):
            c += 1
            for j in range(i+1, len(node_list)):
                n1 = node_list[i]
                n2 = node_list[j]
                if len(n1.domains) == len(n2.domains):
                    if not n1.domains.difference(n2.domains):
                        n1.sameDomainNode.add(n2)
                        n2.sameDomainNode.add(n1)
        self.ete3_tree = tree

    def create_domain_entries(self):
        def initialize_domain_entries():
            domain_entries = {}
            domains = set([h.domain for e in self for h in e.hmmr])
            for d in domains:
                domainObj = Domain(d, set(), set(), set())
                domain_entries[d] = domainObj
            return domain_entries
        domain_entries = initialize_domain_entries()
        for e in self:
            for h in e.hmmr:
                domain_entries[h.domain].hits.add(h)
                domain_entries[h.domain].proteins.add(e.prot)
                domain_entries[h.domain].taxo.add(e.taxo)
        domain_entries = OrderedDict(sorted(domain_entries.items(),
                                     key=lambda kv: len(kv[1].proteins),
                                     reverse=True))
        self.domain_entries = domain_entries

    def compute_upper_node_and_distance(self, core_domains=[]):
        ncbi = NCBITaxa()
        if not self.domain_entries:
            raise Exception("Compute domain_entries first.")
        if not self.ete3_tree:
            raise Exception("Compute ete3_tree first.")

        for d in self.domain_entries.values():
            if d.name not in core_domains:
                distances = []
                if len(d.taxo) == 1:
                    taxo = list(d.taxo)[0]
                    d.upper_node = self.ete3_tree.search_nodes(name=taxo.taxid)[0]
                    d.mean_distance = 0
                else:
                    list_taxids = list(set([t.taxid for t in d.taxo]))
                    domain_tree = ncbi.get_topology(list_taxids)
                    traverse_generator = domain_tree.traverse()
                    d.upper_node = next(traverse_generator)
                    for i in range(len(list_taxids)):
                        for j in range(i+1, len(list_taxids)):
                            dist = self.ete3_tree.get_distance(list_taxids[i], list_taxids[j])
                            distances.append(dist)
                    d.mean_distance = mean(distances)

    def create_domain_graph(self, core_domains):

        def get_vertex_size(nb_domain, interval, min_size):
            if nb_domain == 1:
                return min_size
            else:
                size = min_size+interval*(nb_domain-1)
                return size

        def get_edge_size(nb_occ, interval):
            return nb_occ*interval

        dic_edges = {}
        dic_nb_domain = {}
        all_domains = set()
        for p in self:
            domains = set([h.domain for h in p.hmmr])
            for cd in core_domains:
                domains.discard(cd)
            related_domains = domains.copy()
            for d in domains:
                if d not in core_domains:
                    if d not in dic_nb_domain:
                        dic_nb_domain[d] = 0
                    dic_nb_domain[d] += 1
                related_domains.remove(d)
                all_domains.add(d)
                for d2 in related_domains:
                    edge = tuple(sorted((d, d2)))
                    if edge not in dic_edges:
                        dic_edges[edge] = 0
                    dic_edges[edge] += 1

        list_edges = []
        list_weight = []
        for e in dic_edges:
            list_edges.append(e)
            list_weight.append(dic_edges[e])

        g = Graph()
        g.add_vertices(len(all_domains))
        g.vs["name"] = list(all_domains)
        g.vs["label"] = g.vs["name"]
        g.add_edges(list_edges)
        for vertex in g.vs:
            vertex["weight"] = dic_nb_domain[vertex["name"]]
            vertex["size"] = get_vertex_size(vertex["weight"], 2, 5)
        for e in g.es:
            source = [v for v in g.vs if v.index == e.source][0]
            target = [v for v in g.vs if v.index == e.target][0]
            edge_tuple = tuple(sorted((source["name"], target["name"])))
            nb_occ = dic_edges[edge_tuple]
            min_domains = min(source["weight"], target["weight"])
            e["weight"] = nb_occ/min_domains
            e["label"] = round(nb_occ/min_domains, 2)
            e['width'] = e["weight"]*5

        return g

    def separate_seq_into_fragments(self):
        for e in self:
            e.set_annotated_domains_fragments()
            e.set_Nter_UR_fragment()
            e.set_Cter_UR_fragment()
            e.set_helix_fragments()
            e.set_loop_fragments()

    def add_neighborhood_clusters(self, cluster_tsv):
        cluster_nb = 0
        browse_representative = set()
        f = open(cluster_tsv, "r")
        for l in f:
            l_split = l.rstrip().split("\t")
            representative = l_split[0].split(" ")[0].strip('"')
            seq = l_split[1].split(" ")[0].strip('"')
            prot = seq.split("+")[0]
            seq_name = seq.split("+")[1]
            if representative not in browse_representative:
                cluster_nb += 1
                browse_representative.add(representative)

            if not hasattr(self.entries[prot], "neighborhood_clusters"):
                self.entries[prot].neighborhood_clusters = {}

            self.entries[prot].neighborhood_clusters[seq_name] = cluster_nb
        f.close()

        for e in self:
            if not hasattr(e, "neighborhood_clusters"):
                e.neighborhood_clusters = {}

    def neighborhood_common_matrix(self):
        def get_common_size(list1, list2):
            common_size = 0
            for n in list1:
                if n in list2:
                    common_size += 1
            return common_size

        matrix = {}
        for i in range(100):
            for j in range(i+1, 100):
                neighbors_cluster_list_1 = self[i].get_neighborhood_clusters_number()
                neighbors_cluster_list_2 = self[j].get_neighborhood_clusters_number()
                if not self[i].prot in matrix: 
                    matrix[self[i].prot] = {}
                matrix[self[i].prot][self[j].prot] = get_common_size(neighbors_cluster_list_1, neighbors_cluster_list_2)

        return matrix            
                

class Topology():
    def __init__(self, prot, hmmr, tmhmm, fasta, taxo=None, uniprot_entry=None,
                 annotated_domains_fragments=None, Nter_UR_fragment=None,
                 Cter_UR_fragment=None, helix_fragments=None,
                 loop_fragments=None):
        self.prot = prot
        self.hmmr = hmmr
        self.tmhmm = tmhmm
        self.fasta = fasta
        self.taxo = taxo
        self.uniprot_entry = uniprot_entry
        self.annotated_domains_fragments = annotated_domains_fragments
        self.Nter_UR_fragment = Nter_UR_fragment
        self.Cter_UR_fragment = Cter_UR_fragment
        self.helix_fragments = helix_fragments
        self.loop_fragments = loop_fragments

    def get_taxo(self, function_get_taxid):
        ncbi = NCBITaxa()
        taxname = None
        taxrank = None
        taxid = function_get_taxid(self)
        taxname_dic = ncbi.get_taxid_translator([taxid])
        if taxname_dic:
            taxname = taxname_dic[int(taxid)]
            taxrank_dic = ncbi.get_rank([taxid])
            if taxrank_dic:
                taxrank = taxrank_dic[int(taxid)]
        self.taxo = Taxo(taxid, taxname, taxrank)

    def set_annotated_domains_fragments(self):
        annotated_domains_fragments = []
        for hit in self.hmmr:
            dic_core_domains = {}
            if hit.domain in dic_core_domains:
                raise Exception("Several hits for domain. Handle this part")
            dic_core_domains = {'name': hit.domain, 'seq': hit.sequence,
                                'start': hit.start, 'end': hit.end}
            annotated_domains_fragments.append(dic_core_domains)
        annotated_domains_fragments.sort(key=lambda r: r["start"])
        self.annotated_domains_fragments = annotated_domains_fragments

    def set_Nter_UR_fragment(self):
        dic = {'name': 'N-ter_UR', 'start': 1}
        Nter = self.tmhmm.fragments[0]
        if Nter.end > self.annotated_domains_fragments[0]["start"]:
            dic["end"] = self.annotated_domains_fragments[0]["start"]
        else:
            dic["end"] = Nter.end
        dic["seq"] = self.fasta.get_subsequence(dic["start"], dic["end"])
        self.Nter_UR_fragment = dic

    def set_Cter_UR_fragment(self):
        dic = {'name': "C-ter_UR"}
        Cter = self.tmhmm.fragments[-1]
        dic["end"] = Cter.end
        if Cter.start < self.annotated_domains_fragments[-1]["end"]:
            dic["start"] = self.annotated_domains_fragments[-1]["end"]
        else:
            dic["start"] = Cter.start
        dic["seq"] = self.fasta.get_subsequence(dic["start"], dic["end"])
        self.Cter_UR_fragment = dic

    def set_helix_fragments(self):
        list_fragments = []
        helixes = [f for f in self.tmhmm.fragments if f.cellular_location == "TMhelix"]
        helix_number = 1
        for h in helixes:
            dic = {'name': "TMhelix_" + str(helix_number), 'start': h.start,
                   'end': h.end}
            helix_number += 1
            dic["seq"] = self.fasta.get_subsequence(dic["start"], dic["end"])
            list_fragments.append(dic)
        self.helix_fragments = list_fragments

    def set_loop_fragments(self):
        list_fragments = []
        loops = [f for f in self.tmhmm.fragments[1:-1] if f.cellular_location != "TMhelix"]
        count_inside = 0
        count_outside = 0
        for l in loops:
            dic = {}
            if l.cellular_location == "inside":
                count_inside += 1
                dic["name"] = "inside_loop_" + str(count_inside)
                dic["start"] = l.start
                dic["end"] = l.end
            elif l.cellular_location == "outside":
                count_outside += 1
                dic["name"] = "outside_loop_"+str(count_outside)
                dic["start"] = l.start
                dic["end"] = l.end
            dic["seq"] = self.fasta.get_subsequence(dic["start"], dic["end"])
            list_fragments.append(dic)
        self.loop_fragments = list_fragments

    def set_uniprot_xref(self, uColl):
        p_id = self.prot.split("|")[1]
        try:
            uniprot_entry = uColl.get(p_id)
            self.uniprot_xref = uniprot_entry.xref
        except ValueError as ve:
            ve = str(ve)
            if ve != "Error, empty xmlHandler":
                raise Exception()
            self.uniprot_xref = None

    def get_genome_features(self, enaColl, **kwargs):
        # print("GET GENOME FEATURES")
        # print(kwargs)
        if not self.uniprot_xref:
            raise Exception("Get uniprot xref first.")
        if not self.uniprot_xref.get("EMBL"):
            raise Exception("No EMBL xref. Handle this.")

        # For now, take first id
        ena_id = list(self.uniprot_xref["EMBL"].keys())[0]
        if len(ena_id) != 8: 
            ena_id = ena_id[:6]
        return enaColl.get(ena_id, force_reading_cache=True, **kwargs)

    def set_all_genome_features(self, enaColl, **kwargs):
        self.genome_ena_entry = self.get_genome_features(enaColl, type="all_genome", **kwargs)
        self.genome_ena_id = list(self.uniprot_xref["EMBL"].keys())[0]

    def set_neighborhood(self, number_neighbors, enaColl):
        def filter_type(feature, **kwargs):
            type = kwargs.get("type")
            if not type: 
                raise Exception("Give type argument to filter_type()")
            if feature.type == type:
                return True
            return False

        def filter_no_pseudogene(feature):
            if feature.info.get("other_qualifiers") and "pseudo" in feature.info["other_qualifiers"]:
                return False
            return True

        # Check if required attributes exists
        if not hasattr(self, "uniprot_xref") or not self.uniprot_xref:
            raise Exception("Search uniprot xref first. set_uniprot_xref() method.")
        if not hasattr(self, "genome_ena_entry") or not self.genome_ena_entry:
            raise Exception("Search all genome ena entry first. set_all_genome_features() method")

        # Filter CDS
        # print("filter CDS")
        cds = self.genome_ena_entry.filter(filter_type, type="CDS").features

        # Find protein
        ena_protein_ref = self.uniprot_xref["EMBL"][self.genome_ena_id]
        protein_feature = [f for f in cds if ena_protein_ref in f.info.get("protein_id", [])]
        if not protein_feature:
            raise Exception(self.prot, "protein not found in genome")
        if len(protein_feature) > 1:
            raise Exception(self.prot, "More than 1 protein with id has been found. Handle this")
        protein_feature = protein_feature[0]
        protein_index = cds.index(protein_feature)

        # Get neighborhood
        start = protein_index - number_neighbors
        if start < 0 : 
            start = 0
        end = protein_index + number_neighbors + 1
        if end > len(cds) : 
            end = len(cds)    
        neighborhood = cds[start:end]

        # Correct neighborhood if not in same_contig and delete anchor protein
        neighborhood = [n for n in neighborhood if n.source == protein_feature.source]
        #print(neighborhood)
        neighborhood.remove(protein_feature)

        # Create filters for relaunch genome feature getting
        filter_info = {}
        for n in neighborhood:
            if not n.info.get("locus_tag"):
                # Test with protein_id
                if not n.info.get("protein_id"):
                    raise Exception("No locus_tag, no_protein_id for", self.prot, "Handle this.")
                if "protein_id" not in filter_info:    
                    filter_info["protein_id"] = []
                filter_info["protein_id"] += n.info["protein_id"]
            else:
                if "locus_tag" not in filter_info:
                    filter_info["locus_tag"] = []
                filter_info["locus_tag"] += n.info["locus_tag"]
        # Get genome features again, just keep neighbors and this time keep their sequences.
        # print("NEIGHBORHOOD FEATURES")
        self.set_neighborhood_features(enaColl, info_filter=filter_info, keep_sequence=True, type_filter=["CDS"])

        # Filter pseudogenes
        self.neighborhood_ena_entry = self.neighborhood_ena_entry.filter(filter_no_pseudogene)

    def set_neighborhood_features(self, enaColl, **kwargs):
        self.neighborhood_ena_entry = self.get_genome_features(enaColl, type="neighbors", **kwargs)

    def delete_all_genome_features(self):
        self.genome_ena_entry = None

    @property
    def neighbors_sequences_fasta(self):
        fasta_str = ''
        if not hasattr(self, "neighborhood_ena_entry"):
            raise Exception("Compute neighborhood features first")
        for f in self.neighborhood_ena_entry.features:
            if not f.info.get("translation", None):
                raise Exception("Don't have sequence")
            fasta_str += ">" + self.prot + "+" + f.name + "+" + ";".join(f.info.get("protein_id",'')) + " " + ";".join(f.info.get("product", "")) + "\n"
            fasta_str += f.info.get("translation", [])[0] + "\n"
        return fasta_str

    def get_neighborhood_clusters_number(self):
        if not hasattr(self, "neighborhood_clusters"):
            raise Exception("Compute neighborhood clusters first. Method add_neighborhood_clusters() from TopologyContainer")    

        return list(self.neighborhood_clusters.values())


class Domain():
    def __init__(self, name, hits, proteins, taxo):
        self.name = name
        self.hits = hits
        self.proteins = proteins
        self.taxo = taxo
        self.upper_node = None
        self.mean_distance = None


class Taxo():
    def __init__(self, taxid, taxname, taxrank):
        self.taxid = taxid
        self.taxname = taxname
        self.taxrank = taxrank



import re, io
import gzip

reInstance = re.compile('(\# hmm([\S]+) :: search ([\S]+) against a ([\S]+) database.+?(?=\[ok\])\[ok\])', re.DOTALL)
reType = re.compile('^\# hmm([\S]+) :: search ([\S]+) against a ([\S]+) database[\s]*$')
# hmmsearch :: search profile(s) against a sequence database
reQuery = re.compile('Query:[\s]+([\S].*)[\s]+\[[LM]=[\d]+\]')
reStart = re.compile('^Scores for complete sequences \(score includes all domains\):')
reSum = re.compile('(([\d\.e-]+[\s]+){7})([0-9]+)[\s]+([\S]+)[\s]+([\S].*)$')
reInclude = re.compile('------ inclusion threshold ------')
reDetail = re.compile('Domain annotation for each ([\S]+) \(and alignments\):')
reStop = re.compile('Internal pipeline statistics summary')
reAliLongName = re.compile('^>>[\s]*([\S]+.*)$')
reDomainDetail = re.compile('  == domain [\d]+(?!.* == domain [\d])')
reAlignDef = re.compile('^[\s]*[\d]+[\s]*[\S][\s]+([\S]+)[\s]+([\S]+)[\s]+([\S]+)[\s]+([\S]+)[\s]+([\S]+)[\s]+([\S]+)[\s]+[\.\[\]]{2}[\s]+([\S]+)[\s]+([\S]+)[\s]+[\.\[\]]{2}[\s]+([\S]+)[\s]+([\S]+)[\s]+[\.\[\]{2}[\s]+([\.\d]+)')
reBothStringLetter = re.compile('^([\s]+([\S]+)[\s]+([\S]+)[\s]+)([\S]+)[\s]+([\S]+)')
reBothStringSymbol = re.compile('^([\s]+)([\S]+)')
reEmpty = re.compile('^[\s]*$')
reStrandsLine = re.compile("^([\s]*([\S]+)[\s]+([\d]+)[\s])([\S]+)[\s]([\d]+)[\s]*$")

#reInstance = re.compile('(# hmmsearch.+?(?=\[ok\])\[ok\])', re.DOTALL)

def parse(*inputFiles):
    """
    Parse one or several hmmr results file and store them into container. 

    :param *inputFiles: One or several path to input files, comma-separated
    :return: hmmrContainerFactory.Container
    """

    upType = None
    bigBuffer = ''
    mainContainer = Container()
    for inputFile in inputFiles:
        if inputFile:
            fnOpen = open
            t = 'r'
            if inputFile.endswith('gz'):
                fnOpen = gzip.open
                t = 'rt'

            try:
                f = fnOpen(inputFile, t)
            except IOError:
                print ("Could not read file:", inputFile)
                return Container()

            with f:
                for l in f:
                    bigBuffer += l

                    #print(l)

                runReports = reInstance.findall(bigBuffer)
                #print(runReports)
                if not runReports:
                    raise ValueError("Could not parse upper level in provided hmm(scan/search) file")

                for d in runReports:
                    mainContainer.addParsing(Container(input=io.StringIO(d[0]), upType=d[1]))
    return mainContainer


class Container(object):
    """
    Container that stores hmmr results under different data structure

    :param hmmrEntries: 
    :param dIndex:
    :param pIndex:
    """

    def __init__(self, input=None, upType=None):
       # self.queryHmmFile=None             #../../../data/fad_binding.hmm
       # self.targetSequenceDatabase=None   #./Trembl_50.fasta
       # self.query=None                    #PF08022_full  [M=104]
        self.upType = upType
        if not input:
            self.dIndex={}
            self.pIndex={}
            self.hmmrEntries=set()
        else:
            self.dIndex,self.pIndex,self.hmmrEntries = _parseBuffer(input)


    def addParsing(self, other):
        if self.upType != other.upType:
            if self.upType == None:
                self.upType = other.upType
            elif other.upType == None:
                other.upType = self.upType
        if self.upType != other.upType:
            raise TypeError('Adding', self.upType,' and ', other.upType)

        #self.summary.extend(other.summary)
        #self.details.extend(other.details)
        self.updateParsing(other)
        return self

    def updateParsing(self,other):
        for d in other.dIndex:
            if d not in self.dIndex:
                self.dIndex[d]=set()
            self.dIndex[d].update(other.dIndex[d])

        for p in other.pIndex:
            if p not in self.pIndex:
                self.pIndex[p]=set()
            self.pIndex[p].update(other.pIndex[p])

        self.hmmrEntries.update(other.hmmrEntries)

    def addIndex(self,dic,index,match):
        if index not in dic:
            dic[index]=set()
        dic[index].add(match)

    def __iter__(self):
        '''Iter through matches'''
        for d in self.dIndex:
            for m in self.dIndex[d]:
                yield m

    def add(self,*matches):
        for match in matches:
            prot_id=match.aliShortID
            dom_id=set([hit.hmmID for hit in match.data])
            if len(dom_id)>1:
                raise Exception("dom_id can't have length > 1. Check your code.")
            dom_id=dom_id.pop()
            self.addIndex(self.pIndex,prot_id,match)
            self.addIndex(self.dIndex,dom_id,match)

    def filterProteins(self,fPredicat,**kwargs):
        new_container=Container()
        for protein in self.pIndex:
            matches=self.pIndex[protein]
            if fPredicat(matches,**kwargs):
                new_container.add(*matches)
        return new_container

def _parseBuffer(input):
    dIndex={}
    pIndex={}
    hmmrEntries=set()
    inclusionBool = True
    readBool = False
    detailBool = False
    detailBuffer = []
    summaryBool = False

    queryID = None
    emptyLinesInRow = 0

    for l in input:
       # print(l)
        if not queryID:
            m = reQuery.search(l)
            if m:
                queryID = m.groups()[0]

        if reStart.search(l):
            readBool = True
            continue

        if re.search('\-\-\- full sequence \-\-\-   \-\-\- best 1 domain \-\-\-    \-#dom\-', l):
            #print ("UP")
            inclusionBool = True
            continue

        if reDetail.search(l):
            readBool = False
            detailBool  = True
            continue
        #if reStop.search(l):
        #    break

        if reEmpty.match(l):
            emptyLinesInRow += 1
            if emptyLinesInRow > 1:
               # print('tweet')
                readBool = False
                detailBool = False
        else:
            emptyLinesInRow = 0

        if detailBool:
            if l.startswith(">>"):
                detailBuffer.append(l)
            else :
                if len(detailBuffer) == 0:
                    if re.search('No targets detected that satisfy reporting thresholds', l):
                        print ('No hit found')
                        return ([],[])
                    continue
                detailBuffer[-1] += l

    if not queryID:
        raise ValueError('Could not find query identifier in header')
    #print(detailBuffer)
    dIndex[queryID]=set()
    for rawData in detailBuffer:
        #print(rawData)
        match=Match(rawData,queryID)
        subjctID=match.aliShortID
        if not subjctID:
            continue
        if subjctID not in pIndex:
            pIndex[subjctID]=set()
        pIndex[subjctID].add(match)
        dIndex[queryID].add(match)
        #print(match.data)
        if match.data:
            for hit in match.data:
                domain=hit.hmmID
                prot=hit.aliID
                hmmrEntries.add(HMMObj(prot,domain,hit))
        else:
            hmmrEntries.add(HMMObj(subjctID,None,None))
    #details = [ Match(rawData, queryID) for rawData in detailBuffer ]
    return (dIndex,pIndex,hmmrEntries)

class Match (object):
    def __init__(self, bufferString, queryID):
        self.aliLongID = None
        self.aliShortID = None
        self.hmmID = None
        self.queryID = queryID
        self.parseDetailEntry(bufferString)

    #def __hash__(self):
    #    return hash(self.aliLongID + self.hmmID)

    def parseDetailEntry(self, bufferString):
        buff = bufferString.split('  == domain ')
        self.data = []
        if len(buff) == 1:
            print ("Warning:: " + buff[0])
            self.aliLongID=buff[0].split("\n")[0].lstrip(">> ")
            self.aliShortID=self.aliLongID.split(" ")[0]
            self.data =[]
            return

        for u in buff[0].split('\n'):
            m = reAliLongName.search(u)
            if m:
                self.aliLongID = m.groups()[0]

            m = reAlignDef.search(u)
            if m :
                self.data.append( Alignment(m.groups()) )

        if len( self.data ) != len(buff[1:]):
            raise ValueError('uneven alignment definitions (' + str(len(self.data)) + ') and details (' + str(len(buff[1:])) + ')\n\n' + str(buff))

        for alignmentObject, rawAli in zip( self.data, buff[1:]):
            alignmentObject.parse(rawAli)

        self.aliShortID =  self.data[0].aliID
        self.hmmID = self.data[0].hmmID


    def __repr__(self):
        return str( self.__dict__ )

    def _repr_html_(self):
        htmlContent = ''
        for d in self.data:
            htmlContent += '<table style="border:none">'
            for name, x, y, field in [ (self.hmmID, d.hmmFrom, d.hmmTo, d.hmmStringLetters), ("", "", "", d.matchString), (self.aliShortID, d.aliFrom, d.aliTo, d.aliStringLetters) ]:
                htmlContent += '<tr style="border:none"><td style="border:none">' + name + '</td>' \
                                + '<td style="border:none; color:red; font-size:0.8em">'+ x + '</td><td style="border:none; color:red; font-size:0.8em">' + y + '</td>' \
                                + '<td align="center" style="min-width:20px; text-align:center;border:1px solid rgba(0, 0, 0, .05);">' \
                                + '</td><td align="center" style="min-width:20px; text-align:center;border: 1px solid rgba(0, 0, 0, .05);">'.join(field) +  '</td></tr>'
            htmlContent += '</table>'

        return htmlContent

class Alignment(object):

    def __init__(self, args):

        self.hmmID = None
        self.aliID = None
        self.header = None
        self.score   = args[0]
        self.bias    = args[1]
        self.cEvalue = args[2]
        self.iEvalue = args[3]
        self.hmmFrom = args[4]
        self.hmmTo   = args[5]
        self.aliFrom = args[6]
        self.aliTo   = args[7]
        self.envFrom = args[8]
        self.envTo   = args[9]
        self.acc     = args[10]

        #self.hmmStringSymbols = ''
        self.hmmStringLetters = ''
        self.matchString      = ''
        self.aliStringLetters = ''
        self.hmmSymbolStuff = {}
        self.aliSymbolStuff = {}


    def parse(self, bufferString):

        buf = [ l for l in bufferString.split("\n") ] # if l != ''
        self.header = buf.pop(0)

        #print ('Parsing this:', buf)


        hitLines = [ i for i,l in enumerate(buf) if reStrandsLine.match(l) ]
       # print (hitLines)
       # print(len(hitLines))



        for i in range(0, len(hitLines), 2):
            # Update offset values
            m = reStrandsLine.match(buf[ hitLines[i] ])
            n = reStrandsLine.match(buf[ hitLines[i + 1] ])
            self.aliID = n.groups()[1]
            self.hmmID = m.groups()[1]

#            print ('strand parsing', m.groups())
            offset = len(m.groups()[0])
            length = len(m.groups()[3])
            #print('Line offset is', offset, 'char chunk length is', length)
            reMatchString = r".{" + str(offset) + r"}(.{" + str(length) + r"})"
            mMatch = re.match(reMatchString, buf[ hitLines[i] + 1 ])
            if not mMatch:
                raise ValueError("Could not parse alignment match record", buf[ hitLines[i] + 1 ], "with", reMatchString)
            self.matchString += mMatch.groups()[0]
            reStuffString = r".{" + str(offset) + r"}(.{" + str(length) + r"}) ([A-Z]{2})"
        # i, j a row number, for i we go up and assign to profile data
        #                    for j we go down and assigne to ali data
            self.hmmStringLetters, self.hmmSymbolStuff = plumbParse(buf, hitLines[i], self.hmmStringLetters, self.hmmSymbolStuff, reStuffString,goingUp=True)
            self.aliStringLetters, self.aliSymbolStuff = plumbParse(buf,  hitLines[i + 1], self.aliStringLetters, self.aliSymbolStuff, reStuffString, goingUp=False)
            #print('---->x', self.hmmStringLetters,'<-----x')

    def _parse(self, bufferString):

        buf = [ l for l in bufferString.split("\n") if l != '' ]
        self.header = buf.pop(0)

        iOffset=5
        if len(buf)%5 != 0:
            if len(buf)%4 != 0:
                raise ValueError('unexpected detail alignment data (n=' +  str(len(buf)) +'):\n\n'+str(buf))
            else:
                iOffset=4

        i = 0
        while i < len(buf):
            m = reBothStringSymbol.match(buf[i])
            if iOffset == 5:
                if m:
                    self.hmmStringSymbols +=  m.groups()[1]
            j = i + 1 if iOffset == 5 else i
            m = reBothStringLetter.match(buf[j])
            stringChop = ''
            if m:
                offset = m.groups()[0]
                self.hmmID = m.groups()[1]
                stringChop = m.groups()[3]
                self.hmmStringLetters += stringChop


            j = i + 2 if iOffset == 5 else i + 1
            blank = ' ' * len(offset)
            my_regex = r"^" + re.escape(blank) + r"(.{"+ str(len(stringChop)) + r"})"
            m = re.match(my_regex, buf[j], re.IGNORECASE)

            if m:
                self.matchString += m.groups()[0]

            j = i + 3 if iOffset == 5 else i + 2
            m = reBothStringLetter.match(buf[j])
            stringChop = ''
            if m:
                self.aliID = m.groups()[1]
                stringChop = m.groups()[3]
                self.aliStringLetters += stringChop
            j = i + 4 if iOffset == 5 else i + 3
            m = reBothStringSymbol.match(buf[j])
            if m:
                offset = m.groups()[0]
                self.aliStringSymbols +=  m.groups()[1]

            i += iOffset

        #if not (len(self.aliStringSymbols) == len(self.hmmStringLetters) == len(self.aliStringSymbols) == len(self.aliStringLetters) == len(self.matchString)):
        #    raise ValueError('uneven alignemnt strings')

    def __repr__(self):
        return str( self.__dict__ )

## Low -level alignment buffer parser
def plumbParse(buffer, index, mainStringAccumulator, stuffContainer, reSymbolStuff, goingUp=True):
    while(buffer[index]):
        line = buffer[index]
        #print (index, '::',line)

        if line == '':
            break;
        m = reStrandsLine.match(line)
        mAlt = re.match(reSymbolStuff, line)
        if m:
        #    print ('ACC', m.groups())
        #    print(mainStringAccumulator)
            mainStringAccumulator += m.groups()[3] # Check index
        #    print('Strand View internal here::', mainStringAccumulator)
        elif mAlt:
            stuffKey = mAlt.groups()[1]
            if not stuffKey in stuffContainer:
                stuffContainer[stuffKey] = ''
            stuffContainer[stuffKey] += mAlt.groups()[0]
        else:
            raise ValueError("Unknown alignment record", line)

        index += -1 if goingUp else 1
       # print ("-->", index)
    return mainStringAccumulator, stuffContainer

class HMMObj():
    """
    Object that stores information about one protein domain hit

    :param prot: 
    :param domain: 
    :param hit:
    """

    def __init__(self,prot,domain,hit):
        self.prot=prot
        self.domain=domain
        self.hit=hit
        self.overlapped_hits=[]

    @property    
    def sequence(self):
        """
        Domain sequence in protein, without indels.
        """
        return self.hit.aliStringLetters.replace("-","").upper()

    @property
    def start(self):
        """
        Domain start position in protein
        """
        return int(self.hit.aliFrom)

    @property
    def end(self):
        """
        Domain end position in protein
        """
        return int(self.hit.aliTo)    

    def is_overlapping(self,other_hit,accept_overlap_size):
        """
        Check if two HMMObj overlaps.

        :param other_hit: 
        :param accept_overlap_size: 

        :return: True if overlapping size is bigger than accept_overlap_size. False elsewhere. 
        """
        start1=self.start
        end1=self.end
        start2=other_hit.start
        end2=other_hit.end
        residues1=set([i for i in range(start1,end1+1)])
        residues2=set([i for i in range(start2,end2+1)])
        if len(residues1.intersection(residues2))>accept_overlap_size:
            return True
        return False

    def reinitialize_overlapped_hits(self):
        self.overlapped_hits=[]


from urllib.request import urlopen
import re
import pyproteins.container.customCollection
import pyproteins.container.Core
import pyproteinsext.pfam as pfam
import json

import gzip

from xml.etree.ElementTree import parse, dump, fromstring, register_namespace, ElementTree, tostring


from os.path import expanduser

PfamEntrySet = None
uniprotEntrySet = None

PfamCache = None

from json import JSONEncoder

import pyproteins.utils.make_json_serializable
#def _default(self, obj):
#    return getattr(obj.__class__, "toJSON", _default.default)(obj)
#
#_default.default = JSONEncoder.default
#JSONEncoder.default = _default


## Returns a list of all keyword in collection, along with the list of uniprotObj featuring them
def keyWordChart(uniprotObjIter, kwType='GO'):
    def kwMapper(obj, _type) :
        if _type == 'GO':
            return obj.GO
        raise TypeError("implement other KW plz")

    kwChart = {}
    for uniprotObj in uniprotObjIter:
        for kwObj in  kwMapper(uniprotObj, kwType):
            if kwObj not in kwChart:
                kwChart[kwObj] = []
            kwChart[kwObj].append(uniprotObj)

    return sorted([ (k,v) for k,v in kwChart.items() ], key=lambda x : len(x[1]), reverse=True)

# Give link to uniprot Collection to allow proxy settings
# and cache setting for it and pfam
def getUniprotCollection ():
    global uniprotEntrySet
    if not uniprotEntrySet:
        uniprotEntrySet = EntrySet()

    return uniprotEntrySet

#def setCache(location):
#    print location
#    uniprotEntrySet.setCache(location=location)

#def proxySetting(**kwargs):
#    proxySetting(**kwargs)

def getPfamCollection ():
    global PfamEntrySet
    if not PfamEntrySet:
        home = expanduser("~")
        PfamCacheDefault = home
        PfamEntrySet = pfam.EntrySet(PfamCacheDefault)

    return PfamEntrySet

def proxySetting(**param):
    pyproteins.container.Core.proxySetting(param)

'''
    TODO Isoform, minimal -> affects the fasta sequence
                 Need to check isoform data xml structure, sequence variant specs of Uniprot
'''


def strip(string):
    subString = re.search(".xml$", string)
    if subString:
        return string.split(".")[0]

    return None

def capture(string):
    subString = re.search("[OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}[0-9]){1,2}", string)
    if subString:
        return subString.group()

    return None

def isValidID(string):
    if re.match("^[OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}[0-9]){1,2}$", string):
        return True
    if re.match("^[A-Z]{3}[0-9]{5}$", string):
        return True    
    return False

class fetchEntries():
    def __init__(self, list):
        self.list = list
        self.entrySet = [Entry(id = id) for id in self.list]
    def __iter__(self):
        for entry in self.entrySet:
            yield entry

''' Collection of uniprot entries
customary cache location is "/Users/guillaumelaunay/work/data/uniprot"
'''
class EntrySet(pyproteins.container.customCollection.EntrySet):

    def __init__(self, **kwargs):
        self.ns = '{http://uniprot.org/uniprot}'

        home = expanduser("~")
        cachePath = home     
        self.isXMLCollection = False
        self.isRedisCollection = False
        if 'streamXML' or 'collectionXML' in kwargs:
            self.isXMLCollection = True
            if 'streamXML' in kwargs:
                self.etree = parse(kwargs['streamXML'])
            elif 'collectionXML' in kwargs:
                if kwargs['collectionXML'].endswith('.gz'):
                    with gzip.open(kwargs['collectionXML'], 'r') as gz:
                        self.etree = parse(gz)
                else:
                    self.etree = parse(kwargs['collectionXML'])

            self.etree_root = self.etree.getroot()
            self.index = None
          #  print(f"==> {type(self.etree_root)} {type(self.etree)} <==")
        
        super().__init__(collectionPath=cachePath, constructor=Entry, typeCheck=isValidID, indexer=strip)
        if 'collectionXML' in kwargs:
            print (f"Acknowledged {len(self)} entries {kwargs['collectionXML']}")
      
    def __len__(self):
        if self.isXMLCollection:
            return len(list([ _ for _ in self ]))
        return super().__len__()

    def keys(self):
        """ Returns uniprot identifiers within collection as a generator """
        if self.isXMLCollection:
            for entry in self.etree_root.findall(f"{self.ns}entry"):
                for alt_acc in entry.findall(f"{self.ns}accession"):
                    yield(alt_acc.text)
        else :
            return super().keys()
    
    def has(self, id):
        if self.isXMLCollection:
            if self.index is None:
                self.index = {}
                for acc in self.keys():
                    self.index[acc] = True
            return id in self.index
        else :
            return super().has()

    def __iter__(self):
        if self.isXMLCollection:
            for entry in self.etree_root.findall(f"{self.ns}entry"):
                uniprotID = entry.find(f"{self.ns}accession").text
                yield Entry(uniprotID, xmlEtreeHandler = entry, xmlNS = self.ns)

    def get(self, uniprotID):
        if self.isXMLCollection:
            for entry in self.etree_root.findall(f"{self.ns}entry"):
                for acc in entry.findall(f"{self.ns}accession"):
                    if acc.text == uniprotID: # entry is the node matching provided UNIPROT accessor
                        return Entry(uniprotID, xmlEtreeHandler = entry, xmlNS = self.ns)
            return None
        else :
            print(f"Looking in XML files/dir collection for {uniprotID}")
            return super().get(uniprotID, xmlNS = self.ns)
        
        
    def serialize(self, ext=''):
        global PfamCache
        print ("serializing uniprot collection")
        super().serialize(ext=ext)
        if PfamCache:
            print ("serializing pfam collection")
            getPfamCollection().serialize(ext=ext)
    
    @property
    def taxids(self):
        taxids = set()
        for e in self:
            taxids.add(e.taxid)
        return list(taxids)

class Entry(pyproteins.container.Core.Container):
    ## Wrap and split kwargs
    # Re-encode etree xpath like search 
    # https://docs.python.org/3/library/xml.etree.elementtree.html#example
    # //div[@id='..' and @class='...]

    def __repr__(self):
        asStr = f"{self.id}:{self.AC}\n" 
        asStr += f"{self.name}:{self.fullName}({self.geneName})\n"
        
        if self.STRING_ID:
            asStr += f"STRING_ID:{self.STRING_ID}\n"
        
        asStr += f"taxid:{self.taxid}:{self.lineage}\n"
        asStr += f"KW:{self.KW}\n"
        asStr += f"GO:{self.GO}\n"
        
        return asStr

    def _buildXpath(self, xpath, **kwargs):
        
        _xpath_ = xpath.replace("/",f"/{self._ns}")
        _xpath_ += ' and '.join( [ f"[@{k}='{v}']" for k,v in kwargs.items() ] )

        if not ( _xpath_.startswith('/') or _xpath_.startswith('./') ) :
            _xpath_ = f"{self._ns}{_xpath_}"

        #print(f"building xpath from {xpath} to {_xpath_}")
        return _xpath_

    def _xmlAttrib(self, key, elem=None):
        if elem is None:
            elem = self.xmlHandler
        return elem.attrib[key]

    def _xmlFind(self, tag, elem=None, **kwargs):
        if elem is None:
            elem = self.xmlHandler
        if len(kwargs.keys()) > 1:
            raise(f"Asking for Too many attributes ({len(kwargs.keys())}) in tag xpath search")

        xpathStr = self._buildXpath(tag, **kwargs) 
        return elem.find(xpathStr)
        
    def _xmlFindAll(self, tag, elem=None, **kwargs):
        if elem is None:
            elem = self.xmlHandler
        
        xpath = self._buildXpath(tag, **kwargs) 
        return elem.findall(xpath)

    def __init__(self, id, baseUrl="http://www.uniprot.org/uniprot/", fetchable= True, fileName=None, xmlEtreeHandler=None, xmlNS=None):
        if not id:
            raise TypeError('identifier is empty')
        super().__init__(id, url=baseUrl + str(id) + '.xml', fileName=fileName)
        #pyproteins.container.Core.Container.__init__(self, id, url=baseUrl + str(id) + '.xml', fileName=fileName)
        
        if not xmlNS is None:
            self._ns = xmlNS

        if not xmlEtreeHandler is None:
            self.xmlHandler = xmlEtreeHandler
        else:
            print("Search for", f"{xmlNS}entry")
            self.xmlHandler = self.getXmlHandler(fetchable=fetchable).find(f"{xmlNS}entry")
        
        if self.xmlHandler is None:
            return None
        
        self.name = self._xmlFind("./name").text
        
        if not self._xmlFind("./protein/recommendedName/fullName") is None:
            self.fullName = self._xmlFind("./protein/recommendedName/fullName").text
        else:
            self.fullName = self.name
        
        self.geneName =  None
        e = self._xmlFind("gene")
        if not e is None:
           self.geneName = self._xmlFind("./name", elem=e).text

        self.STRING_ID = None
        e = self._xmlFind("./dbReference", type="STRING")
        if not e is None:
            self.STRING_ID = self._xmlAttrib('id', elem=e)
        
        self.parseAC()
        self.parseLineage()
        self.parseKW()
        self.parseGO()
        self.parseSequence()
        self.parse_subcellular_location()


# Following oarsing stages are disabled since we got rid of bs4
# Need to port them  to lxml, making use of xpath syntax
    def PARSER_TO_PORT_TO_ETREE(self):

       
        self.parseSse()
        self.Ensembl = self.parseEnsembl()
        self.GeneID = self.parseGeneID()
        self.parseSequence()
        self.parsePDB()
        self.parseMIM()
        self.parseDI()
        self.parseORPHA()
        self.xref = self.get_xref()
        self.parseInterpro()

    def __hash__(self):
        return hash(self.id)

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __eq__ (self, other):
        return self.id == other.id

    def toJSON(self):
        #print ('toJSON')
        #asDict = {}
        #for k in self.__dict__.keys():
        #    if k == 'xmlHandler':
        ##        continue
        #    if k != 'GO':
        #        continue
        #    asDict[str(k)] = getattr(self, k)
        #    print(asDict)

        container = {}
        for k, v in self.__dict__.items():
            if k == 'name':
                container[k] = v
            if k == 'GO':
                container[k] = [ go.__dict__ for go in v ]
            if k == 'id':
                container[k] = v
            if k == 'geneName':
                container[k] = v
            if k == 'fullName':
                container[k] = v
            if k == 'taxid':
                container[k] = v
        return container

    def parseAC(self):
        self.AC = [ e.text for e in  self._xmlFindAll("./accession") ]
          
    def parseLineage(self):
        self.lineage = [ e.text for e in  self._xmlFindAll("./organism/lineage/taxon") ]
        e = self._xmlFind('organism/dbReference', type='NCBI Taxonomy')
        self.taxid = self._xmlAttrib('id', elem=e)
    
    def parseMIM(self):
        self.MIM = []
        for e in self.xmlHandler.find_all("dbReference", type="MIM"):
            if str(e.parent.name) == 'entry':
                self.MIM.append(MimKW(e))

    def parseDI(self):
        self.DI = []
        for e in self.xmlHandler.find_all("disease"):
            self.DI.append(DI(e))

    def parseGO(self):
        self.GO = []
        for e in self._xmlFindAll("dbReference", type="GO"):
            self.GO.append(GoKW(
                self._xmlAttrib('id', elem=e),
                self._xmlAttrib('value', elem = self._xmlFind("property", elem=e, type="term") ),
                self._xmlAttrib('value', elem = self._xmlFind("property", elem=e, type="evidence") )
            ))
        
    def parseORPHA(self):
        self.ORPHA = []
        for e in self.xmlHandler.find_all("dbReference", type="Orphanet"):
            if str(e.parent.name) == 'entry':
                self.ORPHA.append(OrphaKW(e))

    def parsePDB(self):
        self.pdbRef = []
        for e in self.xmlHandler.find_all("dbReference", type="PDB"):
            self.pdbRef.append(PDBref(e))

    def parseDomain(self):
        try:
            self.domains=getPfamCollection().map(uniprotID=self.id)
        except:
            self.domains=[]

        #self.domains = []
        #for e in self.xmlHandler.find_all("feature", type="domain"):
        #    buf = Domain(e, self.id)
        #    if buf.description:
        #        self.domains.append(Domain(e, self.id))
        #for e in self.xmlHandler.find_all("feature", type="repeat"):
        #    buf = Domain(e, self.id)
        #    if buf.description:
        #        self.domains.append(Domain(e, self.id))
        #if not self.domains:
        #    print "No domain data found for " + self.id + ", attempting pfam"
        #    try :
        #        self.domains = getPfamCollection().map(uniprotID=self.id)
        #    except ValueError as msg:
        #        print ("Could not bind uniprot to its pfam ressources reason\n" + str(msg))

    def parseSse(self):
        self.sse = []
        for e in self.xmlHandler.find_all("feature", type="strand"):
            self.sse.append(Sse(e))
        for e in self.xmlHandler.find_all("feature", type="turn"):
            self.sse.append(Sse(e))
        for e in self.xmlHandler.find_all("feature", type="helix"):
            self.sse.append(Sse(e))

    def parseKW(self):
        self.KW = []
        
        for e in self._xmlFindAll("./keyword"):
            self.KW.append( UniprotKW(self._xmlAttrib('id', elem=e), e.text) ) 

    def parseSequence(self):
        self.sequence = self._xmlFind("./sequence").text.replace("\n", "").replace(" ", "")
        #self.sequence = Sequence(self.xmlHandler.find("sequence", {"length" : True}))
    #    pass
    
    def parseEnsembl(self):
        #Search for Ensembl id : ENSGXXXXXXXXXXX
        Ensembl_id = []
        for e in self.xmlHandler.find_all("dbReference", type="Ensembl"):
            for e_gene_id in e.find_all('property',type='gene ID'):
                if e_gene_id["value"] not in Ensembl_id:
                    Ensembl_id.append(e_gene_id["value"])
        return Ensembl_id
        
    def parseGeneID(self):
        GeneID= []
        for e_gene_id in self.xmlHandler.find_all("dbReference", type="GeneID"):
            if e_gene_id["id"] not in GeneID:
                GeneID.append(e_gene_id["id"])
        return GeneID

    def parse_subcellular_location(self):
        self.subcellular_location = []
        for citation in self._xmlFindAll('comment', type="subcellular location"):
            for sl in self._xmlFindAll('subcellularLocation', elem = citation):
                for location in self._xmlFindAll('location', elem = sl):
                    self.subcellular_location.append(location.text)
        # except : 
        #     print("WARNING", self.AC, "subcellular location can't be parsed")
        #     print(citation, sl, location)

        # self.subcellular_location = []
        # citations = self._xmlFindAll("comment", type="subcellular location")
        # for citation in citations:
        #     annotation = self._xmlFind('subcellularLocation/location', elem=citation)
           

        #     # if not annotation:
        #     #     print("no annotation", self.AC, annotation)
        #     # else:
        #     try:
        #         self.subcellular_location.append(annotation.text)
        #     except:
        #         print("WARNING", self.AC, "subcellular location can't be parsed")
        #         self.subcellular_location = ['WARNING']
        
    def get_xref(self):
        # print("GET XREF")
        dic_xref = {'EMBL': {}, 'RefSeq': {}}
        # Search EMBL
        for e in self.xmlHandler.find_all("dbReference", type="EMBL"):
            if str(e.parent.name) == 'entry':
                for e_prot_id in e.find_all('property',type='protein sequence ID'):
                    dic_xref["EMBL"][e["id"]] = e_prot_id["value"]
        # Search RefSeq
        for e in self.xmlHandler.find_all("dbReference", type="RefSeq"):
            if str(e.parent.name) == 'entry':
                for e_prot_id in e.find_all('property',type='nucleotide sequence ID'):
                    dic_xref["RefSeq"][e_prot_id["value"]] = e["id"]
        return dic_xref

    def parseInterpro(self):
        self.Interpro = []
        for e in self.xmlHandler.find_all("dbReference", type = "InterPro"):
            for name in e.find_all("property", type = "entry name"):
                self.Interpro.append(Interpro(e))

    @property
    def fasta(self):
        return '>' + str(self.id) + ' ' + str(self.name) + '\n' + str(self.sequence)

    def peptideSeed(self):
        return {
            'id' : self.id,
            'desc' : self.name,
            'seq' : str(self.sequence)
        }



        #self.parseIsoform()

    def hasKW(self, keyword):
        if keyword.upper() in (kw.id.upper() for kw in self.KW):
            return True
        return False
    
    @property
    def isGOannot(self):
        return not len(self.GO) == 0
        
    def hasGO(self, keyword):
        if keyword.upper() in (kw.id.upper() for kw in self.GO):
            return True
        return False

    def hasMIM(self, keyword):
        if keyword.upper() in (kw.id.upper() for kw in self.MIM):
            return True
        return False

    def hasORPHA(self, keyword):
        if keyword.upper() in (kw.id.upper() for kw in self.ORPHA):
            return True
        return False

    def hasDI(self, keyword):
        if keyword.upper() in (kw.id.upper() for kw in self.DI):
            return True
        return False

    # returns a position information container
    def pos(self, i, lookup=False):
        if  i < 1 or i >= len(self.sequence):
            raise IndexError("sequence index \""  + str(i) + "\" is out of range (" + str(len(self.sequence)) + ")")
        d = self._domainFlyCast(position=i)
        if len(d) > 1:
            raise ValueError("\"" + self.id + "\" lays more than one domain \"" + str(len(d)) +
                             "\" at position " + str(i)  )
        domain="SomeDefault"
        if d:
            domain = d[0]
        else:
            l = self._domainFlyCast(NterLookup=i)
            r = self._domainFlyCast(CterLookup=i)
            if l and r:
                domain = { 'outOfBounds' : 'hinge' }
            elif l:
                domain = { 'outOfBounds' : 'Cter' }
            elif r:
                domain = { 'outOfBounds' : 'Nter' }
            else:
                print ("Not a hinge a Cter or a Nter at " + str(i) + " in " + self.id)
        return Position(i, domain=domain, sse=self._getSse(i), letter=self.sequence[i])


    def _domainFlyCast(self, **kwargs):
        if not self.domains:
            raise ValueError("No domain data available for " + self.id)
            #return None
        if 'position' in kwargs:
            results = [ d._dict for d in self.domains if d.owns(kwargs['position']) ]
            return results
        if 'NterLookup' in kwargs:
            for i in range (int(kwargs['NterLookup']), 0, -1):
                d = self._domainFlyCast(position=i)
                if d:
                    return d#d[0]
        if 'CterLookup' in kwargs:
            for i in range (int(kwargs['CterLookup']), len(self.sequence) + 1):
                d = self._domainFlyCast(position=i)
                if d:
                    return d#d[0]


        return None

    def _getSse(self, position):
        i = int(position)
        if not self.sse:
            return None
        for d in self.sse:
            if d.begin <= i <= d.end:
                return d.type
        return "coil"

# Custom encoder for uniprot entity
class EntryEncoder(json.JSONEncoder):
    def default(self, entryObj):
        if isinstance(entryObj, pyproteins.container.Core.Container):
            container = {}
            for k, v in entryObj.__dict__.items():
                if k == 'name':
                    container[k] = v
                if k == 'GO':
                    container[k] = [ go.__dict__ for go in v ]
                if k == 'id':
                    container[k] = v
                if k == 'geneName':
                    container[k] = v
                if k == 'fullName':
                    container[k] = v
            return container
        # Error
        return json.JSONEncoder.default(self, entryObj)



class Position ():
    def __init__(self, number, **kwargs):
        self.number = number
        self.domain = kwargs['domain'] if 'domain' in kwargs else None

        self.sse = kwargs['sse'] if 'sse' in kwargs else None
        self.letter = kwargs['letter'] if 'letter' in kwargs else None

    def __repr__(self):
        return self.letter + str(self.number) + "(" + str(self.sse) + ") : " + str(self.domain)

class Sequence():
    def __init__(self, e):
        if not e:
            raise ValueError("Cant parse sequence")
        self.string = e.text.replace("\n", "").replace(" ", "")
        self.mass =e._xmlAttrib('mass')
    def __len__(self):
        return len(self.string)
    def __getitem__(self, i):
        if isinstance(i, slice):
            newSlice = slice(i.start - 1, i.stop - 1)
            return self.string[newSlice]
        else:
            if i < 1 or i > len(self):
                raise IndexError(str(i) + " out of range [1-" + self(len(self)) + "]")
            return self.string[i - 1]



    def __repr__(self):
        return self.string


class Domain(object):
    def __init__(self, xmlHandler, id):
        self.begin = [int(e['position']) for e in xmlHandler.find_all('begin')]
        self.end = [int(e['position']) for e in xmlHandler.find_all('end')]
        if len(self.begin) != len(self.end):
            raise ValueError("Number of end/stop differs " + str(self.begin) + "/" + str(self.end))
        if 'description' not in xmlHandler:
            #print "Warning Domain w/in \"" + id + "\" does not feature any description"
            self.description = None
        else:
            self.description = xmlHandler['description']
        self.carriedBy = id

    def __eq__(a, b):
        if a.id != b.id:
            return False
        if a.begin != b.begin:
            return False
        if a.end != b.end:
            return False
        if a.description != b.description:
            return False

        return True

    def __repr__(self):
        b = [ str (self.begin[i]) + "-" + str(self.end[i]) for i,e  in enumerate (self.begin) ]
        return "\"" + self.description + "\"\t" + ",".join(b) + "\n"

    def owns(self, position):
        try :
            j = int(position)
        except ValueError as err:
            print ("Improper amino acid position \"" + str(position) + "\"")
            return False

        for i, e  in enumerate (self.begin):
            if self.begin[i] <= j <= self.end[i]:
                return True

        return False



    @property
    def _dict(self):
        return self.__dict__

class Sse():
    def __init__(self, e):
        self.begin = int(e.find('begin')['position'])
        self.end =  int(e.find('end')['position'])
        self.type = e['type']
    def __repr__(self):
        return "SSE:" + self.type + " " + str(self.begin) + "-" + str(self.end)


class annotTerm:
    def __init__(self):
        pass
    def __hash__(self):
        return hash(str(self.id))
    def __eq__(self, other):
        return hash(self) == hash(other)

class GoKW(annotTerm):
    def __init__(self, id,term, evidence):
        self.id = id
        self.term = term
        self.evidence = evidence
    def __repr__(self):
        return self.id + ":" + self.term + "{" + self.evidence + "}"

class MimKW(annotTerm):
    def __init__(self, e):
        self.id = e['id']
        self.value = e.find('property', type='type')['value']
    def __repr__(self):
        return self.id + ":" + self.value

class OrphaKW(annotTerm):
    def __init__(self, e):
        self.id = e['id']
        self.type = e.find("property")['type']
        self.value = e.find('property')['value']

    def __repr__(self):
        return self.id + ": (" + self.type + ")" + self.value

class DI(annotTerm):
    def __init__(self, e):
        self.id = e['id']
        self.name = e.find('name').string
        acronym = e.find('acronym')
        self.acronym = acronym.string if acronym else 'NA'
        self.description = e.find('description').string
    def __repr__(self):
        return self.id + ":" + self.name + " (" + self.acronym + ") {" + self.description + "}"



class PDBref():
    def __init__(self, e):
        self.id = e['id']
        self.method = e.find('property', type='method')['value'] if e.find('property', type='method') else None
        self.resolution = e.find('property',type='resolution')['value'] if e.find('property', type='resolution') else None
        self.chains = e.find('property',type='chains')['value'] if e.find('property', type='chains') else None

    def __repr__(self):
        string = "PDB_id: " + self.id
        elements = []
        if self.method:
            elements.append("method: " + self.method)
        if self.resolution :
            elements.append("resolution: " + self.resolution)
        if self.chains:
            elements.append("chains: " + self.chains)
        if elements:
            return string + ':{' + ','.join(elements)  + '}'

        return string

class UniprotKW():
    def __init__(self, id, text):
        self.id = id
        self.term = text
    def __repr__(self):
        return self.id + ":" + self.term     

class Interpro():
    def __init__(self,e):
        self.id = e["id"]
        self.getName(e)

    def __repr__(self):
        return self.id + ":" + self.name

    def getName(self,e): 
        for name in e.find_all("property", type = "entry name"):
           self.name = name["value"]


class Genome():
    def __init__(self,xmlHandler):
        self.searchEMBL(xmlHandler)
        self.searchRefSeq(xmlHandler)

    def searchEMBL(self,xmlHandler):
        self.EMBLRef=[]
        self.EMBLProteinRef=[]
        for e in xmlHandler.find_all("dbReference", type="EMBL"):
            if str(e.parent.name) == 'entry':
                self.EMBLRef.append(e['id'])
                for e_prot_id in e.find_all('property',type='protein sequence ID'):
                    self.EMBLProteinRef.append(e_prot_id['value'])

    def searchRefSeq(self,xmlHandler):
        self.RefSeqRef=[]
        self.RefSeqProteinRef=[]
        for e in xmlHandler.find_all("dbReference", type="RefSeq"):
            if str(e.parent.name) == 'entry':
                self.RefSeqProteinRef.append(e['id'])
                for e_prot_id in e.find_all('property',type='nucleotide sequence ID'):
                    self.RefSeqRef.append(e_prot_id['value'])



import pyproteins.container.customCollection
import pyproteins.container.Core
from os.path import expanduser
import re

enaEntrySet = None


def getENACollection():
    '''Initialize empty collection'''
    print("Get ENA Collection")
    global enaEntrySet
    enaEntrySet = EntrySet()
    return enaEntrySet


def strip(string):
    subString = re.search(".embl$|.embl.gz$", string)
    if subString:
        return string.split(".")[0]
    return None

#Just name new type of error to catch it
class FormatError(Exception):
    pass

class EntrySet(pyproteins.container.customCollection.EntrySet):
    """
    Inherits from EntrySet, collection of entries.

    :param hmmrEntries: 
    :param dIndex:
    :param pIndex:
    """
    def __init__(self, **kwargs):
        home = expanduser("~")
        cachePath = home

        super().__init__(collectionPath=cachePath, constructor=Entry,
                         indexer=strip)

    def serialize(self, ext=''):
        """Write entries to cache"""
        print("serializing ena collection")
        super().serialize(ext=ext)


class Entry(pyproteins.container.Core.Container):
    """
    Entry object. Store informations about one ENA genome entry, collect from Genbank file. 

    :param metadata: 
    :param features:
    :param pIndex:
    """
    def __init__(self, id, baseUrl="https://www.ebi.ac.uk/ena/data/view/",
                 fileName=None, ext="&display=txt", charge_features=True,
                 rerun=False, url_id=None, **kwargs):
        if not id:
            raise TypeError('identifier is empty')
        if url_id:
            url = baseUrl + str(url_id) + ext
        else:
            url = baseUrl + str(id) + ext
        super().__init__(id, url=url, fileName=fileName)
        self.rerun = rerun
        self.type = type
        if not rerun:
            self.metadata = self.get_metadata()
        if not charge_features:
            self.features = []
        else:
            self.embl_parsing_features(self.raw, **kwargs)

            # Search assembly embl set if there is no features in classical url
            if len(self.features) <= 1 and not rerun:
                short_id = id[:6]
                first_letters = id[:2].lower()
                new_base_url = "http://ftp.ebi.ac.uk/pub/databases/ena/wgs/public/"+first_letters+"/"
                ext = ".dat.gz"
                fileName = None
                self.__init__(self.id, new_base_url, fileName, ext, rerun=True, url_id=short_id, **kwargs)
    
    def get_metadata(self):
        """Parse genbank file (rawdata) to collect metadata. 
        Metadata collected are Project and Sample ids for now. 
        
        :return: dictionnary {"Project" : project_id, "Sample" : sample_id}
        """
        metadata_store = {"Project": "", "Sample": ""}
        project_re = re.compile("^PR\s{3}Project:([\w]+)")
        sample_re = re.compile("^DR\s{3}BioSample; ([\w]+)")

        if isinstance(self.raw, bytes):
            rawData = self.raw.decode()
        else:
            rawData = self.raw

        for l in rawData.split("\n"):
            project_match = project_re.match(l)
            sample_match = sample_re.match(l)
            if project_match:
                metadata_store["Project"] = project_match.group(1)
            if sample_match:
                metadata_store["Sample"] = sample_match.group(1)
        
        return metadata_store

    def __getitem__(self, key):
        try:
            return [f for f in self.features if f.name == key][0]
        except IndexError:
            raise KeyError(key)

    def embl_parsing_features(self, rawData, **kwargs):
        """Parse genbank file to collect features informations. 
        Complete features attributes.
        """
        def conserve_feature(feature, type_filter, info_filter):
            if type_filter:
                if feature.type not in type_filter:
                    return False
            if info_filter:
                for k in info_filter:
                    if not feature.info.get(k):
                        return False
                    if not set(feature.info[k]).intersection(set(info_filter[k])):
                        return False
            return True

        keep_sequence = kwargs.get("keep_sequence", False)
        type_filter = kwargs.get("type_filter", [])
        info_filter = kwargs.get("info_filter", {})

        list_features = []
        feature_header_re = re.compile("^FT\s{3}([^\s]+)\s+([^\s]+)")
        feature_element_re = re.compile("^FT\s+\/(.+)")
        feature_txt_re = re.compile("^FT\s{19}([^\s /].+)")

        if isinstance(rawData, bytes):
            rawData = rawData.decode()

        if not rawData.startswith("ID"):
            raise FormatError("Not embl format")

        feature = None
        l_nb = 0
        dic_nb = {}
        nb_source = 0
        for l in rawData.split("\n"):
            l_nb += 1

            if l.startswith("FT"):  # Feature line
                test_header = feature_header_re.match(l)
                if test_header:
                    # Store previous feature if exists
                    if feature:
                        if conserve_feature(feature, type_filter, info_filter):
                            list_features.append(feature)
                    header = True
                    current_type = test_header.group(1)
                    location = test_header.group(2)
                    feature = Feature(current_type, location)
                    # Set feature source if it's not source itself
                    if current_type == "source":
                        nb_source += 1
                        if self.rerun:
                            feature.name = "Contig"+str(nb_source)
                        else:
                            feature.name = "Genome_element"+str(nb_source)
                        source = feature
                        dic_nb = {}
                    else:
                        if current_type not in dic_nb:
                            dic_nb[current_type] = 0
                        dic_nb[current_type] += 1
                        feature.source = source
                        feature.name = current_type+str(dic_nb[current_type])+"_"+feature.source.name

                test_element = feature_element_re.match(l)
                if test_element:  # Element line. Example : "FT      \pseudo" or "FT      \protein_id=""
                    header = False
                    element = test_element.group(1).split("=")
                    # Complete feature info dictionnary
                    if len(element) == 1:
                        element_key = "other_qualifiers"
                        to_add = element[0].replace('"', '')

                    elif len(element) > 1:
                        element_key = element[0]
                        to_add = element[1].replace('"', '')

                    if not keep_sequence and element_key == "translation":
                        continue

                    if element_key not in feature.info:
                        feature.info[element_key] = []
                    feature.info[element_key].append(to_add)

                test_txt = feature_txt_re.match(l)  # Line with the following text of the element
                if test_txt:
                    if header:
                        feature.location += test_txt.group(1).replace('"', '')
                    else:
                        if not keep_sequence and element_key == "translation":
                            continue
                        feature.info[element_key][-1] += test_txt.group(1).replace('"', '')

        if conserve_feature(feature, type_filter, info_filter):
            list_features.append(feature)

        self.features = list_features

    def filter(self, fPredicat, **kwargs):
        """Generic features filter function. Just keep features when fPredicat is True
        
        :param fPredicat: function that take at least a feature as argument and return True or False. 
        :param **kwargs: optionnal arguments if needed for fPredicat
        :type fPredicat: function
        :type **kwargs: argument dictionnary
        """
        new_Entry = Entry(self.id, self.id, charge_empty=True,
                          fileName=self.fileName)
        for f in self.features:
            if fPredicat(f, **kwargs):
                new_Entry.features.append(f)
        return new_Entry


class Feature():
    """Feature object. Store information about one feature. 

    :param type: feature type, can be CDS, gene, rRNA...
    :param location: feature location in genome/contig
    :param source: parent of this feature (for example, the genome for complete genome, a contig for uncompleted assemblies)
    :param info: dictionnary with all Genbank informations about this feature. Keys are EMBL qualifiers, for example product, translation, gene... 
    """
    def __init__(self, type, location):
        self.type = type
        self.location = location
        self.source = None
        self.info = {}



from marshmallow import Schema, fields, post_load, INCLUDE
from ...uniprot import isValidID, GoKW# impoort unirpiotID validator

class GoTermSchema(Schema):
    id       = fields.Str()
    term     = fields.Str()
    evidence = fields.Str()
    @post_load
    def make_GoKW(self, data, **kwargs):
        return GoKW(**data)

class EntrySchema(Schema):
    #class Meta:
    #    unknown = INCLUDE
    name = fields.Str()
    id   = fields.Str(validate=isValidID)
    geneName = fields.Str(allow_none=True)
    fullName = fields.Str()
    taxid    = fields.Str() # Int ?
    GO = fields.List(fields.Nested(GoTermSchema))
    @post_load
    def make_uniprotEntryProxy(self, data, **kwargs):
        return ProxyEntry(**data)



class ProxyEntry():
    def __init__(self, id, name, geneName, fullName, taxid, GO):
        self.name     = name
        self.id       = id
        self.geneName = geneName
        self.fullName = fullName
        self.taxid    = taxid
        self.GO = GO
    def toJSON(self): # duplicate of pyproteins.Entry.toJSON ...
        container = {}
        for k, v in self.__dict__.items():
            if k == 'name':
                container[k] = v
            if k == 'GO':
                container[k] = [ go.__dict__ for go in v ]
            if k == 'id':
                container[k] = v
            if k == 'geneName':
                container[k] = v
            if k == 'fullName':
                container[k] = v
            if k == 'taxid':
                container[k] = v
        return container
    
    def hasGO(self):
        return keyword.upper() in (kw.id.upper() for kw in self.GO)
    
    def __repr__(self):
        #asStr = f"{self.id}:{self.AC}\n" 
        asStr = f"{self.id}:AC are not yet serialized\n" 
        asStr += f"{self.name}:{self.fullName}({self.geneName})\n"
        
        #if self.STRING_ID:
        #    asStr += f"STRING_ID:{self.STRING_ID}\n"
        asStr += f"taxid:{self.taxid}"
        #asStr += f"taxid:{self.taxid}:{self.lineage}\n"
        #asStr += f"KW:{self.KW}\n"
        asStr += f"GO:{self.GO}\n"
        
        return asStr


import pyproteins.container.customCollection
import pyproteins.container.Core
import re
from os.path import expanduser


refseqEntrySet = None


def isValidID(string):
    '''TO CHANGE'''
    return True


def strip(string):
    '''PROBABLY TO CHANGE'''
    subString = re.search("^[OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}[0-9]){1,2}", string)
    if subString:
        return subString.group()


def getRefseqCollection():
    global refseqEntrySet
    if not refseqEntrySet:
        refseqEntrySet = EntrySet()

    return refseqEntrySet


class EntrySet(pyproteins.container.customCollection.EntrySet):
    def __init__(self, **kwargs):
        home = expanduser("~")
        cachePath = home
        super().__init__(collectionPath=cachePath, constructor=Entry,
                         typeCheck=isValidID, indexer=strip)


class Entry(pyproteins.container.Core.Container):
    def __init__(self, id, baseUrl="https://eutils.ncbi.nlm.nih.gov/entrez/eutils/", fileName=None):
        if not id:
            raise TypeError('identifier is empty')

        #url = baseUrl+"efetch.fcgi?db=nucleotide&id="+id+"&retmode=xml"
        url = baseUrl+"efetch.fcgi?db=protein&id="+id+"&retmode=xml"

        super().__init__(id, url=url, fileName=fileName)

        self.xmlHandler = self.getXmlHandler()
        if not self.xmlHandler:
            return None

        self.length = int(self.xmlHandler.find("GBSeq_length").text)
        self.parseFeatures()

    def __hash__(self):
        return hash(self.id)

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __eq__(self, other):
        return self.id == other.id

    def parseFeatures(self):
        self.Features = []
        for e in self.xmlHandler.find_all("GBFeature"):
            self.Features.append(Feature(e))

    def searchCDS(self, protein_id):
        features = [f for f in self.Features if f.type == "CDS" and f.qualifiers.get("protein_id", None) == protein_id]
        if len(features) == 1:
            return features[0]
        elif len(features) > 1:
            print("More than one feature. How to handle it ?")
            return None
        else:
            print("Protein not found")
            return None

    def getNeighborhood(self, feature, neighborhood_size):
        '''Get list of feature's neighbors, only CDS, from feature start - neighborhood size to feature end + neighborhood size. Only features completely
         contained in this limits are reported.'''
        def isNeighbor(location):
            if other_start_limit:
                if location.start >= other_start_limit:
                    return True

            if other_end_limit:
                if location.end <= other_end_limit:
                    return True

            if location.start >= start_limit and location.end <= end_limit:
                return True

            return False

        self.neighborhood = []
        if len(feature.locations) > 1:
            self.neighborhood.append("To check")
            return

        start = feature.locations[0].start
        end = feature.locations[0].end

        if start < neighborhood_size:
            start_limit = 0
            other_start_limit = self.length-(neighborhood_size-start)
        else:
            start_limit = start-neighborhood_size
            other_start_limit = None

        if end > self.length-neighborhood_size:
            end_limit = self.length
            other_end_limit = self.length-end
        else:
            end_limit = end+neighborhood_size
            other_end_limit = None

        for f in self.Features:
            if f == feature:
                continue
            if f.type != "CDS":
                continue
            neighbor = True
            for l in f.locations:
                if not isNeighbor(l):
                    neighbor = False

            if neighbor:
                self.neighborhood.append(f)


class Feature:
    def __init__(self, e):
        self.type = e.find("GBFeature_key").text
        self.location = e.find("GBFeature_location").text
        self.parseQualifiers(e)

    def parseQualifiers(self, e):
        self.qualifiers = {}
        for q in e.find_all("GBQualifier"):
            name = q.find("GBQualifier_name").text
            try:
                value = q.find("GBQualifier_value").text
            except AttributeError:
                value = ''
            self.qualifiers[name] = value


class Location:
    def __init__(self, start, end, strand, overlap0):
        self.start = start
        self.end = end
        self.strand = strand
        self.overlap0 = overlap0



#import urllib2
from bs4 import BeautifulSoup
import re
import xml.etree.ElementTree as ET
import json


BIOGRID_KEY="8877654116411665fc885b0fc3014cd2"
BIOGRID_UNIPROT_MAPPER_URL="http://thebiogrid.org/downloads/archives/External%20Database%20Builds/UNIPROT.tab.txt"
'''
BIOGRID_ORDERED_JSON_KEYS=["BIOGRID_INTERACTION_ID", "ENTREZ_GENE_A", "ENTREZ_GENE_B",
"BIOGRID_ID_A", "BIOGRID_ID_B", "SYSTEMATIC_NAME_A", "SYSTEMATIC_NAME_B", "OFFICIAL_SYMBOL_A",
"OFFICIAL_SYMBOL_B", "SYNONYMS_A", "SYNONYMS_B", "EXPERIMENTAL_SYSTEM", "EXPERIMENTAL_SYSTEM_TYPE",
"PUBMED_AUTHOR", "PUBMED_ID", "ORGANISM_A", "ORGANISM_B", "THROUGHPUT", "QUANTITATION",
"MODIFICATION", "PHENOTYPES", "QUALIFICATIONS", "TAGS", "SOURCEDB"]
# Fields description at "http://wiki.thebiogrid.org/doku.php/downloads:biogrid_json"
'''
BIOGRID_ORDERED_JSON_KEYS=["BIOGRID_ID_A", "BIOGRID_ID_B", "ENTREZ_GENE_A", "ENTREZ_GENE_B",
 "SYSTEMATIC_NAME_A", "SYSTEMATIC_NAME_B", "OFFICIAL_SYMBOL_A",
"OFFICIAL_SYMBOL_B", "SYNONYMS_A", "SYNONYMS_B", "EXPERIMENTAL_SYSTEM", "EXPERIMENTAL_SYSTEM_TYPE",
"PUBMED_AUTHOR", "PUBMED_ID", "ORGANISM_A", "ORGANISM_B", "THROUGHPUT", "QUANTITATION",
"MODIFICATION", "PHENOTYPES", "QUALIFICATIONS", "TAGS", "SOURCEDB"]


class BIOGRID_DATUM(object):
    def __init__(self, id, data, mapper):
        self.biogridMapper=mapper
        self.data = data
        self.biogridID = id

    def __repr__(self):
        string = ''
        for field in BIOGRID_ORDERED_JSON_KEYS:
            v = self.data[field] if field in self.data else '-'
            if field.startswith("BIOGRID_ID_"):
                w = self.biogridMapper(biogridId=v)
                v = v if not w else "uniprotkb:" + w
            string = string + str(v) + "\t"
        string += str(self.biogridID)
        return string

    @property
    def species(self):
        return (self.data['ORGANISM_A'], self.data['ORGANISM_B'])

    @property
    def uniprotPair(self):
        pp = self.interactors
        if pp[0][0] == "uniprokb:" and pp[1][0] == "uniprokb:":
            a = pp[0][1]
            b = pp[1][1]
            (a,b) = (b,a) if b < a else (a,b)
            return (a, b)
        return None

    @property
    def interactors(self):

        x = self.data['BIOGRID_ID_A']
        y = self.data['BIOGRID_ID_B']
        u = self.biogridMapper(biogridId=x)
        v = self.biogridMapper(biogridId=y)

        if u:
            x = ("uniprokb:", u)
        else:
            x = ("biogrid:", x)

        if v:
            y = ("uniprokb:", v)
        else:
            y = ("biogrid:", y)
        d = ([x],[y]) ## Overkill but required for ducktyping w/ psq.interactors
        #print d
        #return (x, y)
        return d

class BIOGRID(object):
    def __init__(self, **kwargs):  #mapperUrl=BIOGRID_UNIPROT_MAPPER_URL, uniprotMapFile=None
        self.biogridMapper = BIOGRIDMAPPER()
        self.loadBiogridMapper(**kwargs)
        self.data = {}

    def __iter__(self):
        for k in self.data:
            yield BIOGRID_DATUM(k, self.data[k], self.biogridMapper)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return "\n".join( [str(biogridDatum) for biogridDatum in self ] )


    def getBiomolecules(self, _type="uniprot"):

        moleculeList = []
        if _type == "uniprot":
            #for datum in self:
                #print datum
            #    b = datum.interactors
                #print b

            return list(set([ mol[0][1] for datum in self for mol in datum.interactors if mol[0][0] == "uniprokb:" ]))

    def readFile(self, fileName):
        data = ''

        with open(fileName, 'r') as f:
            data = f.read()

        self.load(data)

    # biogridordered keys is the output sequence of dump method
    def load(self, stream, type="biogridOrderedKeys"):
        if not stream:
            print("You must provide a mitab input")
            return
        bufferStr = []

        for line in stream.split("\n"):
            bufferStr.append(line)
        print(len(bufferStr))
        self.tsvBiogridParser(bufferStr)

    def tsvBiogridParser(self, iBuffer):
        for rec in iBuffer:
            #print "----->"  + rec
            record = rec.split("\t")
            key = record.pop()
            self.data[key] = {} # recover the biogrid interaction identifier as primary key
            print(key)
            for l, y in zip(BIOGRID_ORDERED_JSON_KEYS, record):
             #   print '-->' + y

                if y.startswith('uniprotkb:'):
                    m = re.search('^uniprotkb:(.+)$', y)
                    if m:
                        y = self.biogridMapper(uniprotId = m.group(1))
                self.data[key][str(l)] = y #if y != '-' else None

    def clear(self):
        self.data = {}


    def zQuery(self, **kwargs):
        if 'uniprotA' in kwargs and 'uniprotB' in kwargs:
            hArray =(kwargs['uniprotA'], kwargs['uniprotB'])

            cloneOne = BIOGRID()
            chunks = [hArray[0][x:x+50] for x in xrange(0, len(hArray[0]), 50)]
            for p1 in chunks:
                cloneOne.uniprotQuery(p1)

            cloneTwo = BIOGRID()
            chunks = [hArray[1][x:x+50] for x in xrange(0, len(hArray[1]), 50)]
            for p2 in chunks:
                cloneTwo.uniprotQuery(p2)

            self.data = { x:cloneOne.data[x] for x in cloneOne.data if x in cloneTwo.data }

    def dump(self, file=None):
        if file:
            with open(file, 'w') as f:
                f.write("#" + "\t".join(BIOGRID_ORDERED_JSON_KEYS) + "\n")
                f.write(self.__repr__())
        else:
            return self.__repr__()
    def query(self, uniprotId=None, geneIdA=None, geneIdB=None, species=None):
        if uniprotId:
            self.uniprotQuery(uniprotId)
        elif geneIdA and geneIdB:
            self.genePairQuery(geneIdA, geneIdB)
        elif species:
           self.specieQuery(species)

    def specieQuery(self, taxid):
        # 1st check if organism is valid endopoit
        url = "http://webservice.thebiogrid.org/organisms/?format=json&accesskey=" + BIOGRID_KEY
        taxonDict = self._urlToDict(url)
        taxQueryList = []
        taxid = str(taxid)
        # Treat provide taxid 1st as a ncbi taxon identifier, then as a substring of a litteral name

        if taxid in taxonDict:
            taxQueryList.append(taxid)
        else:
            taxQueryList = [ tx for tx in taxonDict if taxid in taxonDict[tx] ]

        if not taxQueryList:
            print (taxid + ' is not a supported Biogrid specie')
            return
        else :
            print (taxQueryList)
        # get the stuff

        for tx in taxQueryList :
            url = "http://webservice.thebiogrid.org/interactions?taxid=" + str(tx) + "&format=json&accesskey=" + BIOGRID_KEY
            data = self._urlToDict(url)
            #print url
            #print data
            self.data.update(data)

    def genePairQuery(self, geneIdA, geneIdB):
        url = "http://webservice.thebiogrid.org/interactions/?geneList="
        url = url + geneIdA + "|" + geneIdB + "&format=json&accessKey=" + BIOGRID_KEY
        try:
            response = urllib2.urlopen(url)
        except urllib2.HTTPError as error:
            print (url + "\nHTTP ERROR " + str(error.code))
            return None
        except urllib2.URLError as error:
            print (url + "\n" + str(error.reason))
            return None
        raw = response.read()
        response.close()
        self._parse(raw)
        self._filter(geneIdA, geneIdB)
        #http://webservice.thebiogrid.org/interactions/?searchNames=true&geneList=cdc27|apc1|apc2&evidenceList=Affinity Capture-MS|Two-hybrid&accesskey=[ACCESSKEY]


    # Returns another instance of BG with filter data elements
    def filter(self, **kwargs):
        target = BIOGRID()
        buf = kwargs['uniprot']
        # For now we dont look into alias uniprot identifiers
        if 'uniprot' in kwargs:
            if isinstance(kwargs['uniprot'], list):
                buf = set(kwargs['uniprot'])
            elif isinstance(kwargs['uniprot'], basestring):
                buf = set([kwargs['uniprot']])
#            elif isinstance(kwargs['uniprot'], set):

            for bgData in self:
                up = bgData.uniprotPair
                if not up:
                    continue
                if set(up) & buf :
                    #print str(up) + ' <===> ' + str(buf)
                    target.data.update({bgData.biogridID : self.data[bgData.biogridID]})

        return target



    def _filter(self, idA, idB):
        dataBuffer = {}
        for interKey in self.data:
            datum = self.data[interKey]
            #print self.data[interKey]
            if idA not in [datum["SYNONYMS_A"], datum["SYNONYMS_B"], datum["SYNONYMS_A"], datum["SYSTEMATIC_NAME_B"], datum["SYSTEMATIC_NAME_A"], datum["OFFICIAL_SYMBOL_A"], datum["OFFICIAL_SYMBOL_B"]]:
                continue
            if idB not in [datum["SYNONYMS_A"], datum["SYNONYMS_B"], datum["SYNONYMS_A"], datum["SYSTEMATIC_NAME_B"], datum["SYSTEMATIC_NAME_A"], datum["OFFICIAL_SYMBOL_A"], datum["OFFICIAL_SYMBOL_B"]]:
                continue
            dataBuffer[interKey] = datum
        self.data = dataBuffer

    def uniprotQuery(self, uniprotId,includeInteractors=True):
        url = "http://webservice.thebiogrid.org/interactions/?geneList="
        qString = None
        if isinstance(uniprotId, list):
            buf = []
            for x in uniprotId:
                biogridId = self.biogridMapper(uniprotId=x)
                if not biogridId:
                    continue
                buf.append(biogridId)
            qString = '|'.join(buf)
        else:
            qString = self.biogridMapper(uniprotId=uniprotId)

        if not qString:
                return

        incBool = 'true' if includeInteractors else 'false'
        url = url + qString + "&searchbiogridids=true&includeInteractors=" + incBool + "&format=json&accessKey=" + BIOGRID_KEY

        data = self._urlToDict(url)
        self.data.update(data)

    def _urlToDict(self,url):

        try:
            response = urllib2.urlopen(url)
        except urllib2.HTTPError as error:
            print (url + "\nHTTP ERROR " + str(error.code))
            return None
        except urllib2.URLError as error:
            print (url + "\n" + str(error.reason))
            return None
        raw = response.read()
        response.close()

        buf = json.loads(raw)
        if isinstance(buf, list):
            if len(buf) == 0:
                return {}
            raise TypeError('unexpected list on json parsing' + str(buf))
        # We expected dict w/ unique key per interactions
        return buf

    def loadBiogridMapper(self, uniprotMapFile=None):

        if uniprotMapFile:
            with open(uniprotMapFile,'r') as f:
                print ('reading uniprot mapping from ' + uniprotMapFile)
                data = f.read()
                self._loadMapper(data)
            return

        print ('reading uniprot mapping from ' + BIOGRID_UNIPROT_MAPPER_URL)
        try:
            response = urllib2.urlopen(BIOGRID_UNIPROT_MAPPER_URL)

        except urllib2.HTTPError as error:
            print ("loadBioGridMapper url resolution failed")
            print ("HTTP ERROR " + str(error.code))
            return None
        except urllib2.URLError as error:
            print ("loadBioGridMapper url resolution failed")
            print (error.reason)
            return None
        raw = response.read()
        response.close()
        self._loadMapper(raw)


    def _loadMapper(self, stream):
        self.biogridMapper.load(stream)
    def analyse(self):
        if len(self) == 0: return None
        container = { "pmids" : [], "experiments" : [], "experimentTypes" : [],
                     "highThroughput" : False, "lowThroughput" : False }
        for k, record in self.data.iteritems():
            if record["PUBMED_ID"] not in container["pmids"]:
                container["pmids"].append(record["PUBMED_ID"])
                container["experiments"].append(record["EXPERIMENTAL_SYSTEM"])
                container["experimentTypes"].append(record["EXPERIMENTAL_SYSTEM_TYPE"])
                if "Low Throughput" in record["THROUGHPUT"]: container["lowThroughput"] = True
                if "High Throughput" in record["THROUGHPUT"]: container["highThroughput"] = True
        return container
    def getExperimentalSystems(self):

        registredSystem  = []
        for key in self.data:
            if self.data[key]["EXPERIMENTAL_SYSTEM"] not in registredSystem:
                registredSystem.append(self.data[key]["EXPERIMENTAL_SYSTEM"])

        return registredSystem

class BIOGRIDMAPPER():
    def __init__(self):
        self.uniprotToBiogrid = {}
        self.biogridToUniprot = {}
    def __call__ (self, uniprotId=None, biogridId=None):
        if biogridId and uniprotId:
            self.add(str(biogridId), str(uniprotId))
            return
        if (biogridId):
            value = self.toUniprot(str(biogridId))
            return value
        if (uniprotId):
            value = self.toBiogrid(str(uniprotId))
            return value

    def read(self, fileName):
        s=''
        with open(fileName, 'r') as f:
            s = f.read()
        self.load(s)

    def load(self, stream):
        for line in stream.split("\n"):
            if line.startswith("#") or not line:continue
            array = line.split()
            self(uniprotId=array[0], biogridId=array[1])

    def toUniprot (self, id):
        if id in self.biogridToUniprot:
            return self.biogridToUniprot[id]
        #print "Biogrid mapper failed to convert \"" + str(id) + "\" to a uniprot ID"
        return None

    def toBiogrid (self, id):
        if id in self.uniprotToBiogrid:
            return self.uniprotToBiogrid[id]
        #print "Biogrid mapper failed to convert \"" + str(id) + "\" to a biogrid ID"
        return None

    def add(self, biogridId=None, uniprotId=None):
        self.biogridToUniprot[str(biogridId)] = str(uniprotId)
        self.uniprotToBiogrid[str(uniprotId)] = str(biogridId)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import, print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup

setup(
  name = 'pyproteinsext',
  version = '0.5',
  license='BSD',
  description = 'Extending pyproteins for bioinformatics tools&services',
  author = 'Guillaume Launay, Juliette Martin et al',
  author_email = 'pitooon@gmail.com, juliette.martin@ibcp.fr',
  url = 'https://github.com/glaunay/pyproteinsExt', # use the URL to the github repo
  packages=find_packages('src'),
  package_dir={'': 'src'},
  include_package_data=True,
  zip_safe=False,
  py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
  download_url = 'https://github.com/glaunay/pyproteinsExt/tarball/0.5', # I'll explain this in a second
  keywords = ['protein', 'sequence'], # arbitrary keywords
  classifiers = [],
  install_requires=[
          'pyproteins', 'networkx', 'bs4', 'biopython'
      ],
   package_data = {
   'pyproteinsext': ['external/submitPsipred.sh','static/psicquicRegistryDefault.xml']
   },
  #data_files=[
  #          ('external', ['external/pathos.tar.bz']),
  #          ('bin', ['bin/module1.py']),
  #          ('conf',['conf/confModule1.json'])
  #    ]
  #  dependency_links = [
  #      "http://dev.danse.us/trac/pathos"
  #  ]
)
